#ifndef COMMAND_H
#define COMMAND_H

#include <iostream>
#include <string>
#include <algorithm>
#include <map>
#include <set>

#include <boost/variant.hpp>
#include "boost/tuple/tuple.hpp"

#include <linguisticgraph/linguisticgraph.h>

#include "utils.h"

namespace ogrecomponents{

class Command
{
public:
    Command(std::string name, Verbosity verbose = NONE)
        :m_name(name), m_verbose(verbose){}

    virtual ~Command(){}

    const std::string& name()const{return m_name;}

    virtual bool process(const std::map<unsigned, unsigned>& matching_nodes,
                         const std::map<unsigned, unsigned>& matching_edges,
                    LinguisticGraph *sentence) = 0;

protected:
    std::string m_name;
    Verbosity m_verbose;

};

class AddEdgeCommand: public Command
{
    typedef std::map<std::string, std::string> MapFeatures;
    typedef MapFeatures::const_iterator MapFeaturesConstIter;
    typedef MapFeatures::iterator MapFeaturesIter;

public:
    AddEdgeCommand(unsigned src, unsigned tar,
                   std::string name, Verbosity verbose = NONE)
        :Command(name, verbose), m_src(src), m_tar(tar){}

    ~AddEdgeCommand();

    void add_feature(std::string key, std::string value);

    bool process(const std::map<unsigned, unsigned>& matching_nodes,
                 const std::map<unsigned, unsigned>& matching_edges,
            LinguisticGraph *sentence);

private:
    unsigned m_src, m_tar;
    MapFeatures m_features;
};

class RemoveCommand: public Command
{
public:
    enum Mode{Node = 0, Edge};
    RemoveCommand(unsigned id, Mode mode, std::string name, Verbosity verbose = NONE)
        :Command(name, verbose), m_id(id), m_mode(mode){}

    ~RemoveCommand();

    bool process(const std::map<unsigned, unsigned>& matching_nodes,
                 const std::map<unsigned, unsigned>& matching_edges,
            LinguisticGraph *sentence);

private:
    unsigned m_id;
    Mode m_mode;
};

class MoveEdgeCommand: public Command
{
public:
    typedef std::map<std::string, std::string> MapFeatures;
    typedef MapFeatures::const_iterator MapFeaturesConstIter;
    typedef MapFeatures::iterator MapFeaturesIter;

    enum Type{Move = 0, Clone};
    MoveEdgeCommand(unsigned eid, unsigned new_src, unsigned new_tar, Type type,
                    std::string name, Verbosity verbose = NONE)
        :Command(name, verbose), m_eid(eid),
          m_newSrc(new_src), m_newTar(new_tar), m_type(type)
    {}

    ~MoveEdgeCommand(){}

    void add_feature(std::string key, std::string value);

    bool process(const std::map<unsigned, unsigned>& matching_nodes,
                 const std::map<unsigned, unsigned>& matching_edges,
            LinguisticGraph *sentence);

private:
    void modifyFeatures(LinguisticGraph *graph, unsigned eid);

private:
    unsigned m_eid, m_newSrc, m_newTar;
    MapFeatures m_features;
    Type m_type;
};

class FeatureCommand: public Command
{
public:
    enum Element{Node = 0, Edge};
    enum Mode{Append = 0, Add, Remove};

    typedef boost::variant<
        std::set<std::string>,
        boost::tuples::tuple<unsigned, std::string, Element> > FeatureValues;

    FeatureCommand(unsigned id, std::string key,
                   FeatureValues values,
                   Mode mode, Element element,
                   std::string name, Verbosity verbose = NONE)
        :Command(name, verbose), m_id(id), m_key(key), m_values(values), m_mode(mode), m_element(element)
    {}

    ~FeatureCommand(){}

    bool process(const std::map<unsigned, unsigned>& matching_nodes,
                 const std::map<unsigned, unsigned>& matching_edges,
                 LinguisticGraph *sentence);

private:
    void feature_command_on_edge(LinguisticEdgeProperties &edge,
                                 const std::set<std::string> &values);
    void feature_command_on_node(LinguisticNodeProperties &node,
                                 const std::set<std::string> &values);

private:
    struct Visitor: boost::static_visitor<bool>
    {
        Visitor(const std::map<unsigned, unsigned>& mnodes,
                const std::map<unsigned, unsigned>& medges,
                LinguisticGraph *sent)
            :matching_nodes(mnodes), matching_edges(medges), sentence(sent)
        {}

        std::set<std::string> values()const {return m_values;}

        bool operator()(const std::set<std::string> &values);
        bool operator()(const boost::tuples::tuple<unsigned, std::string, Element> &values);

    private:
        const std::map<unsigned, unsigned>& matching_nodes;
        const std::map<unsigned, unsigned>& matching_edges;
        LinguisticGraph *sentence;

        std::set<std::string> m_values;
    };

private:
    unsigned m_id;
    std::string m_key;
    FeatureValues m_values;
    Mode m_mode;
    Element m_element;
};

class ModifyFeatureCommand: public Command
{
public:
    enum Element{Node = 0, Edge};
    ModifyFeatureCommand(unsigned id, std::string key, std::string condition,
                         std::string replacement, Element element,
                         std::string name, Verbosity verbose = NONE)
        :Command(name, verbose), m_id(id), m_key(key), m_condition(condition), m_replace(replacement),
          m_element(element){}

    bool process(const std::map<unsigned, unsigned>& matching_nodes,
                 const std::map<unsigned, unsigned>& matching_edges,
            LinguisticGraph *sentence);

private:
    void modify_on_node(LinguisticNodeProperties &node);
    void modify_on_edge(LinguisticEdgeProperties &edge);

private:
    unsigned m_id;
    std::string m_key;
    std::string m_condition;
    std::string m_replace;

    Element m_element;

};

class StatusEdgeCommand: public Command
{
public:
    StatusEdgeCommand(unsigned id, bool surfacic, bool deep,
                      std::string name, Verbosity verbose = NONE)
        :Command(name, verbose), m_id(id), m_surfacic(surfacic), m_deep(deep){}

    bool process(const std::map<unsigned, unsigned>& matching_nodes,
                 const std::map<unsigned, unsigned>& matching_edges,
            LinguisticGraph *sentence);

private:
    unsigned m_id;
    bool m_surfacic, m_deep;
};

}

#endif //COMMAND_H
