#ifndef COMPILER_H
#define COMPILER_H

#include <linguisticgraph/linguisticgraph.h>
#include <grql/grql.hpp>
#include<boost/tuple/tuple.hpp>

#include "utils.h"

#include "rule.h"

namespace ogrecomponents{

class Compiler
{
    typedef string::const_iterator iterator_type;
    typedef grql::parser::Skipper<iterator_type> skipper_type;

public:
    typedef std::vector<Rule*>::const_iterator RulesConstIter;
    enum Mode{Text = 0, File};

    Compiler(std::string content, Mode mode, Verbosity verbose = NONE)
        :m_content(content), m_mode(mode), m_verbose(verbose){}

    ~Compiler();

    bool compile();

    std::pair<RulesConstIter, RulesConstIter> rules()const
    {return make_pair(m_rules.begin(), m_rules.end());}

protected:
    template<typename AST>
    pair<string, bool> string_parse(string &source, AST& ast)
    {
        iterator_type iter = source.begin();
        iterator_type end = source.end();
        grql::error_handler<iterator_type> error_handler(iter, end);
        grql::parser::Skipper<iterator_type> skipper;
        grql::parser::rule_parser<iterator_type, skipper_type> parser(error_handler);
        bool success = boost::spirit::qi::phrase_parse(iter, end, parser, skipper, ast);
        return make_pair(string(iter, end), success);
    }

private:
    std::string m_content;
    Mode m_mode;
    Verbosity m_verbose;
    std::vector<Rule*> m_rules;
};

struct RuleMaker
{
    RuleMaker(vector<Rule*> &rules, Verbosity verbose = NONE)
        :m_rules(rules), m_verbose(verbose){}

    bool process_rule(const grql::ast::rule_t &rule);
    bool process_subrule(const grql::ast::subrule_t &subrule, string name);
    bool process_graph(const grql::ast::graph_t &graph, Rule *rule);
    bool process_nac(const grql::ast::graph_t &nac_graph, Rule *rule);
    bool process_precedence(const grql::ast::precedence_t &precedence, Rule *rule);
    bool process_command(const grql::ast::command_t &command, Rule *rule);
    bool process_constraint(const grql::ast::constraint_t &constraint, Rule *rule);

private:
    vector<Rule*> &m_rules;
    Verbosity m_verbose;
};

struct GraphMaker: boost::static_visitor<bool>
{
    enum Mode{Normal, NAC};

    GraphMaker(Rule *rule, Verbosity verbose = NONE)
        :m_rule(rule), m_graph(NULL), m_verbose(verbose), m_mode(Normal){}


    bool process_graph(const grql::ast::graph_t &graph,
                       int parent = -1);
    std::pair<bool, unsigned> process_node(const grql::ast::node_t &node);
    bool process_edge(const grql::ast::edge_t &edge);

    LinguisticGraph *graph()const{return m_graph;}

    void setMode(Mode m){m_mode = m;}

    std::pair<std::set<unsigned>, std::set<unsigned> > nac_ids()
    const{return make_pair(nac_nids, nac_eids);}

private:
    void merge_node_feature(LinguisticNodeProperties &final,
                            const LinguisticNodeProperties &toMerge);
    void merge_edge_feature(LinguisticEdgeProperties &final,
                            const LinguisticEdgeProperties &toMerge);

private:
    Rule *m_rule;
    LinguisticGraph *m_graph;
    Verbosity m_verbose;
    Mode m_mode;

    std::set<unsigned> nac_nids, nac_eids;

    std::pair<LinguisticEdgeProperties, std::string> pending_edge;
};

struct DFAMaker: boost::static_visitor<bool>
{
    DFAMaker(DFA *dfa = NULL);

    bool process_dfa(const std::vector<grql::ast::transition_t> &dfa);
    bool process_transition(const grql::ast::transition_t &transition);

    DFA *dfa()const {return m_dfa;}

private:
    DFA *m_dfa;
};

struct FeatureStructureMaker: boost::static_visitor<bool>
{
    enum Type{Node = 0, Edge};
    FeatureStructureMaker(Type t)
        :m_type(t), m_var(""){m_node = NULL; m_edge = NULL; m_nac = NULL;}

    bool operator()(const grql::ast::var_feature_t &var_feature);
    bool operator()(const grql::ast::feature_struct_t &feature);

    void node_feature(const grql::ast::feature_struct_t &feature);
    void edge_feature(const grql::ast::feature_struct_t &feature);   

    std::string variable()const {return m_var;}

    LinguisticNodeProperties *m_node;
    LinguisticEdgeProperties *m_edge;
    NAC *m_nac;

private:
    Type m_type;
    std::string m_var;
};

struct CommandMaker: boost::static_visitor<bool>
{
    CommandMaker(Rule *rule, Verbosity verbose = NONE)
        :m_command(NULL), m_rule(rule), m_verbose(verbose){}

    bool operator()(const grql::ast::remove_command_t & remove_cmd);
    bool operator()(const grql::ast::add_edge_t & add_edge_cmd);
    bool operator()(const grql::ast::move_edge_t& move_edge_cmd);
    bool operator()(const grql::ast::feature_command_t& feature_cmd);
    bool operator()(const grql::ast::modify_command_t& feature_modify);
    bool operator()(const grql::ast::status_edge_t &status_cmd);

private:
    Command *m_command;
    Rule *m_rule;
    Verbosity m_verbose;

private:
    struct FeatureValuesVisitor: boost::static_visitor<bool>
    {
        FeatureValuesVisitor(Rule *rule, Verbosity verbose)
            :m_rule(rule), m_verbose(verbose){}

        FeatureCommand::FeatureValues values()const {return m_values;}

        bool operator()(const grql::ast::feature_dict_t &feature_dict);
        bool operator()(const std::vector<std::string> &simple_values);

    private:
        FeatureCommand::FeatureValues m_values;
        Rule *m_rule;
        Verbosity m_verbose;
    };
};


struct ConstraintMaker: boost::static_visitor<bool>
{
    ConstraintMaker(Rule *rule, Verbosity verbose = NONE)
        :m_rule(rule), m_constraint(NULL), m_verbose(verbose){}

    bool operator()(const grql::ast::move_constraint_t &move);
    bool operator()(const grql::ast::simple_constraint_t &simple);
    bool operator()(const grql::ast::feature_constraint_t &relabel);

private:
    Rule *m_rule;
    Constraint *m_constraint;
    Verbosity m_verbose;

private:

    std::map<std::string, std::string> features( const std::string & status);

    struct VarMaker: boost::static_visitor<bool>
    {
        VarMaker(Rule *rule)
          :m_rule(rule){}

        bool operator()(const grql::ast::cvar_edge_t &var_nodes);
        bool operator()(const std::string &var_edge);

        Rule::ConstraintIdType id()const{return m_eid;}

    private:
        Rule *m_rule;
        Rule::ConstraintIdType m_eid;
    };
};

}

#endif
