#ifndef CONSTRAINTS_H
#define CONSTRAINTS_H

#include "linguisticgraph/linguisticgraph.h"
#include "ogrecomponents/dfa.h"
#include "ogrecomponents/utils.h"

#include <string>
#include <set>

#include <boost/algorithm/string.hpp>

using namespace std;

namespace ogrecomponents{

class DFA;

class Constraint
{
public:
    Constraint();
    virtual ~Constraint(){}

    virtual bool trigger(LinguisticGraph *g, unsigned eid) = 0;

    void resetHistoric(){m_historic.clear();}

protected:
    std::set<unsigned> m_historic;

protected:
    void evaluateStatus(const std::string &status, std::string &original, std::string &newone)
    {
        if( status.empty() )
            return;

        vector<string> vec;
        boost::algorithm::split( vec, status, boost::algorithm::is_any_of(":"),
                                 boost::algorithm::token_compress_on);

        if(vec.size() == 2)
        {
            original = vec[0];
            newone   = vec[1];
        }else if(vec.size() == 1){
            original = "";
            newone   = vec[0];
        }
    }

    bool isCanonical(string& label)
    {
        if(label.begin() != label.end())
        {
            if(*label.begin() == ':')
            {
                label.erase(0, 1);
                return true;
            }
        }
        return false;
    }

    bool isBoth(const string &label, string &final, string &canonical)
    {
        vector<string> vec;
        boost::algorithm::split( vec, label, boost::algorithm::is_any_of(":"),
                                 boost::algorithm::token_compress_on );

        if(vec.size() == 2)
        {
            final = vec[0];
            canonical = vec[1];
            return true;
        }
        return false;
    }

    bool isFinal(const string &label)
    {
        return boost::algorithm::find_first(label, ":");
    }

    bool hasEdge(unsigned src, unsigned tar, unsigned edge,
                 LinguisticGraph *g, bool testTarget)
    {
        //Get edge properties
        LinguisticEdgeProperties &edgeProp = g->edgeProperties(edge);

        if(testTarget)
        {
            //Browse out edges from source
            LinguisticGraph::out_edge_iter beout, eeout;
            for(boost::tie(beout, eeout) = g->outEdges(src); beout != eeout; ++beout)
            {
                //Edge properties of source out edges
                LinguisticEdgeProperties &prop = g->edgeProperties(*beout);
                unsigned test = g->nodeProperties(g->targetNode(*beout)).id;
                if(tar == test && m_historic.find(prop.id) != m_historic.end())
                {
                    return true;
                }
            }
        }
        else
        {
            //Browse in edges from source
            LinguisticGraph::in_edge_iter bein, eein;
            for(boost::tie(bein, eein) = g->inEdges( g->nodeProperties(src).id );
                bein != eein; ++bein)
            {
                //Edge properties of source out edges
                LinguisticEdgeProperties &prop = g->edgeProperties(*bein);
                unsigned test = g->nodeProperties(g->sourceNode(*bein)).id;
                if(src == test && m_historic.find(prop.id) != m_historic.end() )
                {
                    return true;
                }
            }
        }
        return false;
    }

    void fix_label(LinguisticEdgeProperties &prop, string type){
        if(type == "f")
            prop.setRelCanonical("");
        else if(type == "c")
            prop.setRelFinal("");
    }

    void insert_features(const std::map<std::string, std::string> &features,
                         LinguisticEdgeProperties &edge){

        std::map<std::string, std::string>::const_iterator it = features.begin();
        for(; it != features.end(); ++it){
            edge.setFeature( it->first, it->second );
        }
    }
};


class SimpleConstraint: public Constraint
{
public:
    SimpleConstraint(set<string> labels, std::map<std::string, std::string> features,
                     Verbosity verbose);

    ~SimpleConstraint();

    virtual bool trigger(LinguisticGraph *g, unsigned eid) = 0;


protected:
    set<string> m_labels;
    std::map<std::string, std::string> m_features;
    Verbosity m_verbose;

protected:
    set< pair<unsigned, string> > solutions(LinguisticGraph *g, set<unsigned> edges);
};

/*class RedirectUp: public SimpleConstraint
{
public:
    RedirectUp(set<string> labels,
               Verbosity verbose, Logic l = Exists);
    bool trigger(GraphWrapper *&g, string eid);
};*/

class RedirectDown: public SimpleConstraint
{
public:
    RedirectDown(set<string> labels, std::map<std::string, std::string> features,
                 Verbosity verbose);
    ~RedirectDown(){}

    bool trigger(LinguisticGraph *g, unsigned eid);
};

class ShareUp: public SimpleConstraint
{
public:
    ShareUp(set<string> labels, std::map<std::string, std::string> features,
            Verbosity verbose);
    bool trigger(LinguisticGraph *g, unsigned eid);
};

class ShareDown: public SimpleConstraint
{
public:
    ShareDown(set<string> labels, std::map<std::string, std::string> features,
              Verbosity verbose);
    ~ShareDown(){}

    bool trigger(LinguisticGraph *g, unsigned eid);
};

class AddFeature : public SimpleConstraint
{
public:
    AddFeature(set<string> labels, string final, string canonical,
               std::map<std::string, std::string> features, Verbosity verbose);
    ~AddFeature(){}

    bool trigger(LinguisticGraph *g, unsigned eid);

private:
    string m_final, m_canonical;
};

/*class MoveUp: public Constraint
{
public:
    MoveUp(DFA *dfa, Verbosity verbose);
    bool trigger(GraphWrapper *&g, string eid);

protected:
    DFA *m_dfa;
    Verbosity m_verbose;
};*/

class AutomatonConstraint: public Constraint
{
public:
    enum Direction{DOWN, UP};

    AutomatonConstraint(DFA *dfa, string status, Verbosity verbose);
    virtual ~AutomatonConstraint(){}

protected:
    DFA *m_dfa;
    string m_status;
    Verbosity m_verbose;

protected:
    bool inspectDFA(LinguisticGraph *graph, set<unsigned> & ids,
                    map<unsigned, string> &solutions, Direction dir);
};

class MoveDown: public AutomatonConstraint
{
public:
    MoveDown(DFA *dfa, string status, Verbosity verbose);
    ~MoveDown();

    bool trigger(LinguisticGraph *g, unsigned eid);
};

class MoveShareDown: public AutomatonConstraint
{
public:
    MoveShareDown(DFA *dfa, string status, Verbosity verbose);
    ~MoveShareDown();

    bool trigger(LinguisticGraph *g, unsigned eid);
};

}

#endif // CONSTRAINTS_H
