#ifndef DFA_H
#define DFA_H

#include <boost/tr1/unordered_map.hpp>
#include <boost/tr1/unordered_set.hpp>
#include <string>
#include <exception>

using namespace std;
using namespace boost;
using namespace unordered;

namespace ogrecomponents{

class DFAError: public std::exception
{
public:
    DFAError(const std::string& what="") throw()
        :m_what("DFA exception: "+what)
    {}

    virtual const char* what() const throw()
    {
        return m_what.c_str();
    }

    virtual ~DFAError() throw()
    {}

private:
    string m_what;
};

class AlphabetError: public std::exception
{

public:
    AlphabetError(const std::string& what="") throw()
        :m_what("DFA alphabet exception: "+what)
    {}

    virtual const char* what() const throw()
    {
        return m_what.c_str();
    }

    virtual ~AlphabetError() throw()
    {}

private:
    string m_what;
};


class DFA
{
public:
    typedef unordered_map<unsigned, unordered_map<string, unsigned> > Transitions;
    typedef Transitions::iterator TransitionsIterator;
    typedef Transitions::const_iterator TransitionsConstIterator;

    /**
     * @brief DFA
     * @param alpha : set of possible labels
     */
    DFA(unordered_set<string> alpha);

    /**
     * @brief addTransition
     * @param src source state
     * @param input input string
     * @param dest target state
     *
     * This function add a new transition to the automaton :
     *  src - input -> dest
     *
     *input could be of the following forms:
     * 1. "label"
     * 2. "label1|label2|...|labelx"
     * 3. "__not__label"
     * 4. "__not__label1|label2|...|labelx"
     *
     * __not__ is a special token which is used to
     * say that the transition should not recognise all of the
     * labels after __not__.
     */
    void addTransition(unsigned src, string input, unsigned dest);

    /**
     * @brief addFinalState
     * @param state state number
     *
     * This function adds a final state to the automaton.
     */
    void addFinalState(unsigned state);

    /**
     * @brief isFinal
     * @param state
     * @return true if the state number is final, otherwise false
     */
    bool isFinal(unsigned state) const;

    /**
     * @brief next
     * @param input : label to be recognised
     * @return the next state regarding of the input
     * @warning This function makes the DFA jump to another state if possible
     */
    unsigned next(string input);

    /**
     * @brief testTransition
     * @param input
     * @return the next state regarding of the input
     * It's just a test, so the DFA does not jump
     */
    unsigned testTransition(string input);

    //bool recognize(string word);

    inline friend std::ostream& operator<<(std::ostream& out, const DFA& dfa)
    {
        for(TransitionsConstIterator it = dfa.transitions.begin(); it != dfa.transitions.end(); ++it)
        {
            bool final = dfa.isFinal(it->first);
            out << it->first << (final ? "(f)" : "") << endl;
            unordered_map<string, unsigned> end = it->second;
            for(unordered_map<string, unsigned>::const_iterator it2 = end.begin(); it2 != end.end(); ++it2)
            {
                final = dfa.isFinal(it2->second);
                out << "\t" << "- " << it2->first << " -> " << it2->second << (final ? "(f)" : "") << endl;
            }
        }

        return out;
    }

    /**
     * @brief start
     * @return the start state number
     */
    int start()const{return startState;}

    /**
     * @brief current
     * @return the current state number
     */
    unsigned current()const{return currentState;}

    /**
     * @brief notToken
     * @return const string corresponding to the __not__ special token
     * @see isNot
     */
    const std::string notToken()const {return "__not__";}

    /**
     * @brief reset
     * Reset the DFA to initial state
     */
    void reset() {currentState = startState;}

private:
    unsigned nextState(unsigned src, string input);

    bool isNot(string input);
    bool isDisjunction(string input);

    string notLabel(string input);
    vector<string> disjunct(string input);
    string removeLogical(string input);

    //void createAlphabet();
private:
    int startState;
    unsigned currentState;
    Transitions transitions;
    unordered_set<unsigned> finalStates;

    unordered_set<string> alphabet;
};

}

#endif // DFA_H
