#include "ogrecomponents/compiler.h"
#include "ogrecomponents/nac.h"

#include "linguisticgraph/format/conll.h"

namespace ogrecomponents{

using namespace std;

Compiler::~Compiler()
{
    for(std::vector<Rule*>::iterator it = m_rules.begin(); it != m_rules.end(); ++it)
    {
        Rule *r = *it;
        delete r;
    }

    m_rules.clear();
}

bool Compiler::compile()
{
    vector<grql::ast::rule_t> parsed_rules;
    pair<string, bool> result = string_parse(m_content, parsed_rules);

    if(result.second)
    {
        for(int i = 0; i < parsed_rules.size(); i++)
        {
            RuleMaker rulemaker(m_rules, m_verbose);
            if(!rulemaker.process_rule(parsed_rules[i]))
                return false;          
        }
        return true;
    }
    else
        return false;
}

////////////////////////////////////
// Rule maker
////////////////////////////////////

bool RuleMaker::process_rule(const grql::ast::rule_t &rule)
{
    if(m_verbose >= DEBUG)
        std::cerr << "Compilation for rule " << rule.name_ << std::endl;

    bool result = process_subrule(rule.rule_, rule.name_);
    return result;
}

bool RuleMaker::process_subrule(const grql::ast::subrule_t &subrule, std::string name)
{
    //Compute each match
    for(int i = 0; i < subrule.matchs_.size(); i++)
    {
        Rule *rule = new Rule;
        rule->setName(name);

        if( !process_graph(subrule.matchs_[i], rule) )
            return false;

        // Compute each NAC on graph
        // WARNING to match the NACs with the matches
        // we NEED to compute them after the subgraphs
        // because subgraphs will instantiate the variables
        // and when NACs will be created, they will get
        // exactly the same ID for the subgraph rule and the NAC
        // which will allow the software to decide if a NAC is applicable
        // or not
        for(int i = 0; i < subrule.nacs_.size(); i++)
        {
            if( !process_nac(subrule.nacs_[i], rule) )
                return false;
        }

        //WARNING need to be process after match
        //so that we know each variable
        if( subrule.precedence_ ){
            if( !process_precedence(*subrule.precedence_, rule) )
               return false;
        }

        //Compute each command
        if(subrule.commands_)
        {
            grql::ast::commands_t commands = *subrule.commands_;
            for(int i = 0; i < commands.size(); i++)
            {
                if( !process_command(commands[i], rule) )
                    return false;
            }
        }

        //Compute each constraint
        if(subrule.constraints_)
        {
            grql::ast::constraints_t constraints = *subrule.constraints_;
            for(int i = 0; i < constraints.size(); i++)
            {
                if( !process_constraint(constraints[i], rule) )
                    return false;
            }
        }

        m_rules.push_back(rule);
    }    

    return true;
}

bool RuleMaker::process_graph(const grql::ast::graph_t &graph, Rule *rule)
{
    GraphMaker gm(rule, m_verbose);
    bool result = gm.process_graph(graph);

    if(result){
        /*std::string content = "";
        CoNLL::save(gm.graph(), content);
        std::cerr << content << std::endl;*/
        rule->setGraph(gm.graph());
    }

    return result;
}

bool RuleMaker::process_nac(const grql::ast::graph_t &nac_graph, Rule *rule)
{
    GraphMaker gm(rule, m_verbose);
    gm.setMode(GraphMaker::NAC); //Put the graph maker object in NAC mode (ie it adds node if it exists)

    bool result = gm.process_graph(nac_graph);

    if(result){      
        NAC *nac = new GraphNAC(gm.graph(), gm.nac_ids());

        /*std::string content = "";
        CoNLL::save(gm.graph(), content);
        std::cerr << "NAC Graph: " << std::endl << content << std::endl;*/

        rule->addNAC(nac);
        return true;
    }

    return false;
}

bool RuleMaker::process_precedence(const grql::ast::precedence_t &precedence, Rule *rule)
{
    std::vector<unsigned> ids;
    for(int i = 0; i < precedence.size(); i++){
        std::string var = precedence[i];
        std::pair<unsigned, bool> var_result = rule->get_node_id(var);

        if(var_result.second)
            ids.push_back(var_result.first);
        else{
            //Unable to find variable, two solutions:
            // 1. Maybe the variable refers to an edge
            if(rule->get_edge_id(var).second){
                std::cerr << "In rule " << rule->name() << ", "
                          << var << " is declared as an edge" << std::endl;
                std::cerr << "In precedence statement, you must use only nodes!" << std::endl;
                return false;
            // 2. Variable is not declared
            }else{
                std::cerr << "In rule " << rule->name() << ", "
                          << var << " is not declared in match" << std::endl;
                std::cerr << "You can't use undeclared variable in precedence" << std::endl;
                return false;
            }
        }
    }

    NAC *nac = new PrecedenceNAC(ids);
    rule->addNAC(nac);

    return true;
}

bool RuleMaker::process_command(const grql::ast::command_t &command, Rule *rule)
{
    CommandMaker cm(rule, m_verbose);
    return boost::apply_visitor(cm, command);
}

bool RuleMaker::process_constraint(const grql::ast::constraint_t &constraint, Rule *rule)
{
    ConstraintMaker cm(rule, m_verbose);
    return boost::apply_visitor(cm, constraint);
}

////////////////////////////////////
// Graph maker
////////////////////////////////////
bool GraphMaker::process_graph(const grql::ast::graph_t &graph, int parent)
{
    if(!m_graph)
        m_graph = new LinguisticGraph;

    std::pair<bool, unsigned> result = process_node(graph.source_);
    if(!result.first)
    {
        std::cerr << "Rule " << m_rule->name() << ": error during node creation..." << std::endl;
        return false;
    }

    if(parent != -1)
    {
        std::string direction = pending_edge.second;
        unsigned src = direction == "-" ? (unsigned)parent : result.second;
        unsigned tar = direction == "-" ? result.second : (unsigned)parent;
        m_graph->addEdge(src, tar, pending_edge.first);
    }

    for(int i = 0; i < graph.rest_.size(); i++)
    {
        const grql::ast::relation_t &relation = graph.rest_[i].get();

        if( !process_edge(relation.edge) )
        {
            std::cerr << "Rule " << m_rule->name() << ": error during edge creation..." << std::endl;
            return false;
        }

        for(int j = 0; j < relation.graph.size(); j++)
        {
            if( !process_graph(relation.graph[j], result.second) )
            {
                std::cerr << "Rule " << m_rule->name() << ": error during subgraph creation" << std::endl;
                return false;
            }
        }
    }

    return true;
}

std::pair<bool, unsigned> GraphMaker::process_node(const grql::ast::node_t &node)
{
    FeatureStructureMaker fsm(FeatureStructureMaker::Node);
    if(!boost::apply_visitor(fsm, node.element))
        return make_pair(false,0);

    if(!fsm.m_node)
        return make_pair(false,0);

    std::string var = fsm.variable();
    LinguisticNodeProperties n(*fsm.m_node);
    bool node_exists = false;

    if(!var.empty())
    {
        if(m_rule->get_edge_id(var).second)
        {
            std::cerr << "In rule " << m_rule->name() << ", "
                      << var << " is declared as an edge" << std::endl;
            std::cerr << "It cannot refer to a node and an edge is the same rule." << std::endl;
            return make_pair(false, 0);
        }

        if(m_rule->get_node_id(var).second)
            node_exists = true;        
    }

    if(node_exists){
        n.id = m_rule->get_node_id(var).first;
        if( m_mode == GraphMaker::NAC ){
            LinguisticNodeProperties &match_node = m_rule->graph()->nodeProperties(n.id);
            merge_node_feature(n, match_node);
            m_graph->addNode(n.id, n);
            nac_nids.insert(n.id);
        }
    }
    else
    {
        if( m_mode == GraphMaker::NAC )
            n.id = m_rule->graph()->lastNodeId()+1;
        else
            n.id = m_graph->nodesSize();
        m_graph->addNode(n.id, n);

        if(!var.empty() && m_mode != GraphMaker::NAC )
            m_rule->add_node_var(var, n.id);

        if(fsm.m_nac){          
            //Handle Negative Application Condition
            FeatureNAC *fnac = dynamic_cast<FeatureNAC*>(fsm.m_nac);

            if(fnac)
                fnac->setId(n.id);
            else{
                std::cerr << "Unabled to find the feature NAC" << std::endl;
                return make_pair(false, 0);
            }
            m_rule->addNAC(fsm.m_nac);
        }
    }

    return make_pair(true, n.id);
}

bool GraphMaker::process_edge(const grql::ast::edge_t &edge)
{
    FeatureStructureMaker fsm(FeatureStructureMaker::Edge);
    if(!boost::apply_visitor(fsm, edge.element_))
        return false;

    if(!fsm.m_edge)
        return false;

    std::string var = fsm.variable();

    LinguisticEdgeProperties current_edge(*fsm.m_edge);
    current_edge.id = m_graph->lastEdgeId()+1;

    if(!var.empty())
    {
        if(m_rule->get_node_id(var).second)
        {
            std::cerr << "In rule " << m_rule->name() << ", " << var << " is declared as an node" << std::endl;
            std::cerr << "It cannot refer to a node and an edge in the same rule." << std::endl;
            return false;
        }

        if( m_mode == GraphMaker::NAC )
        {
            if(m_rule->get_edge_id(var).second){
                current_edge.id = m_rule->get_edge_id(var).first;            

                LinguisticEdgeProperties &match_edge = m_rule->graph()->edgeProperties(current_edge.id);
                merge_edge_feature(current_edge, match_edge);
                nac_eids.insert(current_edge.id);
            }
        }
        else
            m_rule->add_edge_var(var, current_edge.id);
    }

    if(fsm.m_nac){
        //Handle Negative Application Condition
        FeatureNAC *fnac = dynamic_cast<FeatureNAC*>(fsm.m_nac);

        if(fnac)
            fnac->setId(current_edge.id);
        else{
            std::cerr << "Unabled to find the feature NAC" << std::endl;
            return false;
        }
        m_rule->addNAC(fsm.m_nac);
    }    

    pending_edge = std::make_pair(current_edge, edge.direction_);

    return true;
}

void GraphMaker::merge_edge_feature(LinguisticEdgeProperties &final,
                                    const LinguisticEdgeProperties &toMerge)
{
    //Copy the edge in matching
    LinguisticEdgeProperties e(toMerge);

    //If the edge in the NAC has extra feature, add them to e
    if(final.relFinal() != "")
       e.setRelFinal(final.relFinal());

    if(final.relCanonical() != "")
        e.setRelCanonical(final.relCanonical());

    if(final.typeString() != e.typeString())
        e.setDependencyType(final.type());

    //Edge in NAC is now e
    final = e;
}

void GraphMaker::merge_node_feature(LinguisticNodeProperties &final,
                                    const LinguisticNodeProperties &toMerge)
{
    //Copy the node in matching
    LinguisticNodeProperties n(toMerge);

    //If the node in the NAC has extra feature, add them to n
    if(final.token() != "")
        n.setToken( final.token() );

    if(final.pos() != "")
        n.setPos( final.pos() );

    if(final.hyperpos() != "")
        n.setHyperpos( final.hyperpos() );

    if(final.lemma() != "")
        n.setLemma( final.lemma() );

    LinguisticNodeProperties::const_feature_iter fit, end;
    for(boost::tie(fit, end) = final.featuresIterator(); fit != end; ++fit){
        n.addFeature(fit->first, fit->second);
    }

    //Node in NAC is now n
    final = n;
}

////////////////////////////////////
// Feature structure maker
////////////////////////////////////

bool FeatureStructureMaker::operator ()(const grql::ast::var_feature_t &var_feature)
{
    m_var = var_feature.variable_;

    if(var_feature.structure_)
    {
        FeatureStructureMaker fsm(m_type);
        bool result = fsm(*var_feature.structure_);
        m_node = fsm.m_node;
        m_edge = fsm.m_edge;
        m_nac  = fsm.m_nac;
        return result;
    }
    else
    {
        if(m_type == FeatureStructureMaker::Node)
            m_node = new LinguisticNodeProperties;
        else
            m_edge = new LinguisticEdgeProperties;
        return true;
    }
}

bool FeatureStructureMaker::operator ()(const grql::ast::feature_struct_t& feature)
{
    if(m_type == FeatureStructureMaker::Node)
    {
        node_feature(feature);
        return true;
    }
    else
    {
        edge_feature(feature);
        return true;
    }
}

void FeatureStructureMaker::node_feature(const grql::ast::feature_struct_t &feature)
{
    m_node = new LinguisticNodeProperties;   
    for(int i = 0; i < feature.size(); i++)
    {
        grql::ast::extended_feature_t ext_feature = feature[i];
        grql::ast::feature_t feat = ext_feature.feature_;

        if(ext_feature.condition_){
            if(!m_nac)
                m_nac = new FeatureNAC(m_node->id, FeatureNAC::Node);

            FeatureNAC *fnac = dynamic_cast<FeatureNAC*>(m_nac);
            for(int j = 0; j < feat.values_.size(); j++)
            {
                std::string val = feat.values_[j].value_;                
                fnac->addFeature(feat.key_, val);
            }
        }else{

            for(int j = 0; j < feat.values_.size(); j++)
            {
                std::string val = feat.values_[j].value_;
                if(feat.key_ == "token"){
                    m_node->addToken(val);
                }else if(feat.key_ == "lemma"){
                    m_node->addLemma(val);
                }else if(feat.key_ == "hypercat"){
                    m_node->addHyperpos(val);
                }else if(feat.key_ == "cat"){
                    m_node->addPos(val);
                }else{
                    m_node->addFeature(feat.key_, val);
                }
            }
        }
    }
}

void FeatureStructureMaker::edge_feature(const grql::ast::feature_struct_t &feature)
{
    m_edge = new LinguisticEdgeProperties;
    for(int i = 0; i < feature.size(); i++)
    {        
        grql::ast::extended_feature_t ext_feature = feature[i];
        grql::ast::feature_t feat = ext_feature.feature_;

        if(ext_feature.condition_)
        {
            if(feat.key_ == "label")
            {
                if(!m_nac)
                    m_nac = new FeatureNAC(m_edge->id, FeatureNAC::Edge);

                FeatureNAC *fnac = dynamic_cast<FeatureNAC*>(m_nac);
                for(int j = 0; j < feat.values_.size(); j++)
                {
                    std::string val = feat.values_[j].value_;
                    if(linggraph::isBothLabel(val)){
                        std::vector<std::string> parts;
                        boost::algorithm::split(parts, val, boost::algorithm::is_any_of(":"));
                        if(parts.size() == 2){
                            fnac->addFeature(feat.key_, parts[0]);
                            fnac->addFeature(feat.key_, ":"+parts[1]);
                        }else
                            fnac->addFeature(feat.key_, val);
                    }
                    else
                        fnac->addFeature(feat.key_, val);
                }
            }//End of if key_ == label
        }
        else
        {
            for(int j = 0; j < feat.values_.size(); j++)
            {
                if(feat.key_ == "label")
                {
                    std::string label = feat.values_[j].value_;
                    if(linggraph::isFinalLabel(label)){
                        m_edge->addRelFinal(label);
                    }else if(linggraph::isCanonicalLabel(label)){
                        m_edge->addRelCanonical(label);
                    }else if(linggraph::isBothLabel(label)){
                        std::vector<std::string> parts;
                        boost::algorithm::split(parts, label, boost::algorithm::is_any_of(":"));
                        if(parts.size() == 2){
                            m_edge->addRelFinal(parts[0]);
                            m_edge->addRelCanonical(parts[1]);
                        }else{
                            m_edge->addRelFinal(label);
                            m_edge->addRelCanonical("");
                        }
                    }
                }
                else if(feat.key_ == "surfacic")
                {
                    bool surfacic = false;
                    std::string status = feat.values_[j].value_;
                    if( status == "y" )
                        surfacic = true;
                    else
                        surfacic = false;

                    m_edge->setSurfacicState(surfacic);
                }
                else if(feat.key_ == "deep")
                {
                    bool deep = false;
                    std::string status = feat.values_[j].value_;
                    if( status == "y" )
                        deep = true;
                    else
                        deep = false;
                    m_edge->setDeepState(deep);
                }
            } //End of for feat.values_.size()

        }//End of NAC condition test
    }
}


////////////////////////////////////
// DFA maker
////////////////////////////////////
DFAMaker::DFAMaker(DFA *dfa)
    :m_dfa(dfa)
{
    if(!m_dfa){
        m_dfa = new DFA(unordered_set<string>());
    }
}

bool DFAMaker::process_dfa(const std::vector<grql::ast::transition_t> &dfa)
{

    for(int i = 0; i < dfa.size(); i++){
        if( !process_transition(dfa[i]) )
            return false;
    }

    return true;
}

bool DFAMaker::process_transition(const grql::ast::transition_t &transition)
{
    int from_state = transition.from_.first;
    bool from_is_final = transition.from_.second == 'f' ? true : false;
    int to_state   = transition.to_.first;
    bool to_is_final = transition.to_.second == 'f' ? true : false;

    char special_token = transition.input_.first ? *(transition.input_.first) : '_';
    std::string labels = boost::algorithm::join(transition.input_.second, std::string("|"));
    std::string final_input = "";
    switch(special_token){
    case '!':
        final_input = m_dfa->notToken() + labels;
        break;
    default:
        final_input = labels;
    }

    m_dfa->addTransition(from_state, final_input, to_state);

    if(from_is_final)
        m_dfa->addFinalState(from_state);
    if(to_is_final)
        m_dfa->addFinalState(to_state);

    return true;
}


////////////////////////////////////
// FeatureValues visitor
////////////////////////////////////
bool CommandMaker::FeatureValuesVisitor::operator ()(const grql::ast::feature_dict_t &feature_dict)
{
    std::pair<unsigned, bool> result = m_rule->get_node_id(feature_dict.variable_);
    FeatureCommand::Element elt_type = FeatureCommand::Node;
    if( !result.second ){
        result = m_rule->get_edge_id(feature_dict.variable_);
        elt_type = FeatureCommand::Edge;

        if( !result.second ){
            std::cerr << feature_dict.variable_ << " does not name a node or an edge" << std::endl;
            return false;
        }
    }

    //Insert values into boost::variant
    m_values = make_tuple(result.first, feature_dict.value_, elt_type);
    return true;
}

bool CommandMaker::FeatureValuesVisitor::operator ()(const std::vector<std::string> &simple_values)
{
    std::set<std::string> values;
    std::copy(simple_values.begin(), simple_values.end(), std::inserter(values, values.begin()));

    //Insert values into boost::variant
    m_values = values;

    return true;
}

////////////////////////////////////
// Command maker
////////////////////////////////////

bool CommandMaker::operator()(const grql::ast::remove_command_t & remove_cmd)
{
    std::string name = remove_cmd.command_;
    /*if(m_verbose)
        std::cerr << "Command " << name << std::endl;*/

    RemoveCommand::Mode mode;
    if(name == "remove_node")
        mode = RemoveCommand::Node;
    else if(name == "remove_edge")
        mode = RemoveCommand::Edge;
    else{
        std::cerr << "In rule " << m_rule->name() << ", unknown remove command: " << name << std::endl;
        return false;
    }

    std::pair<unsigned,bool> id;

    if(mode == RemoveCommand::Node)
    {
        id = m_rule->get_node_id(remove_cmd.var_);
        if(!id.second)
        {
            std::cerr << "In rule " << m_rule->name() << ", for " << name
                      << ", " << remove_cmd.var_
                      << " does not exist or does not refer to a node" << std::endl;
            return false;
        }
    }
    else
    {
        id = m_rule->get_edge_id(remove_cmd.var_);
        if(!id.second)
        {
            std::cerr << "In rule " << m_rule->name() << ", for " << name
                      << ", " << remove_cmd.var_
                      << " does not exist or does not refer to a node" << std::endl;
            return false;
        }
    }

    m_command = new RemoveCommand(id.first, mode, remove_cmd.command_, m_verbose);    
    m_rule->addCommand(m_command);

    return true;
}

bool CommandMaker::operator()(const grql::ast::add_edge_t & add_edge_cmd)
{
    /*if(m_verbose)
        std::cerr << "Command " << add_edge_cmd.command_ << std::endl;*/

    std::pair<unsigned,bool> src = m_rule->get_node_id(add_edge_cmd.varSrc_),
            tar = m_rule->get_node_id(add_edge_cmd.varTar_);

    if(!src.second || !tar.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", for " << add_edge_cmd.command_ << ", ";
        std::cerr << add_edge_cmd.varSrc_ << " and/or " << add_edge_cmd.varTar_
                  << " do not exist or do not refer to proper node" << std::endl;
        return false;
    }

    m_command = new AddEdgeCommand(src.first, tar.first,
                                   add_edge_cmd.command_, m_verbose);

    AddEdgeCommand *c = dynamic_cast<AddEdgeCommand*>(m_command);

    for(int i = 0; i < add_edge_cmd.features_.size(); i++)
    {
        grql::ast::feature_t feature = add_edge_cmd.features_[i].feature_;
        std::string key = feature.key_;
        for(int j = 0; j < feature.values_.size(); j++)
            c->add_feature(key, feature.values_[j].value_);
    }

    m_rule->addCommand(m_command);

    return true;
}

bool CommandMaker::operator()(const grql::ast::move_edge_t& move_edge_cmd)
{
    /*if(m_verbose)
        std::cerr << "Command " << move_edge_cmd.command_ << std::endl;*/


    std::pair<unsigned,bool> eid = m_rule->get_edge_id(move_edge_cmd.var_);
    if(!eid.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", for " << move_edge_cmd.command_ << ", ";
        std::cerr << move_edge_cmd.var_ << " do not exist or not refer to proper edge" << std::endl;
        return false;
    }

    std::pair<unsigned,bool> newSrc = m_rule->get_node_id(move_edge_cmd.varNewSrc_),
            newTar  = m_rule->get_node_id(move_edge_cmd.varNewTar_);

    if(!newSrc.second || !newTar.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", for " << move_edge_cmd.command_ << ", ";
        std::cerr << move_edge_cmd.varNewSrc_ << " and/or " << move_edge_cmd.varNewTar_
                  << " do not exist or not refer to proper nodes" << std::endl;
        return false;
    }

    MoveEdgeCommand::Type t;
    if(move_edge_cmd.command_ == "clone_edge")
        t = MoveEdgeCommand::Clone;
    else if(move_edge_cmd.command_ == "move_edge")
        t = MoveEdgeCommand::Move;

    m_command = new MoveEdgeCommand(eid.first, newSrc.first, newTar.first, t,
                                    move_edge_cmd.command_, m_verbose);

    MoveEdgeCommand *c = dynamic_cast<MoveEdgeCommand*>(m_command);

    if(move_edge_cmd.features_)
    {
        grql::ast::feature_struct_t features = *move_edge_cmd.features_;
        for(int i = 0; i < features.size(); i++)
        {
            grql::ast::feature_t feature = features[i].feature_;
            std::string key = feature.key_;
            for(int j = 0; j < feature.values_.size(); j++)
                c->add_feature(key, feature.values_[j].value_);
        }
    }

    m_rule->addCommand(m_command);

    return true;
}

bool CommandMaker::operator()(const grql::ast::feature_command_t& feature_cmd)
{
    std::string name = feature_cmd.command_;

    /*if(m_verbose)
        std::cerr << "Command " << name << std::endl;*/

    FeatureCommand::Mode mode;
    if(name == "add_feature")
        mode = FeatureCommand::Add;
    else if(name == "remove_feature")
        mode = FeatureCommand::Remove;
    else if(name == "append_feature")
        mode = FeatureCommand::Append;   
    else
    {
        std::cerr << "In rule " << m_rule->name() << ", command "
                  << name << " does not exits" << std::endl;
        return false;
    }

    FeatureCommand::FeatureValues values;
    if(feature_cmd.values_)
    {
        FeatureValuesVisitor fvm( m_rule, m_verbose );
        boost::apply_visitor( fvm, *feature_cmd.values_ );
        values = fvm.values();
    }

    std::pair<unsigned,bool> id_node = m_rule->get_node_id(feature_cmd.var_);
    std::pair<unsigned,bool> id_edge = m_rule->get_edge_id(feature_cmd.var_);
    if(!id_node.second && !id_edge.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", for " << name << ", ";
        std::cerr << feature_cmd.var_ << " does not exist or not refer to a node or an edge" << std::endl;
        return false;
    }
    unsigned id = id_node.second ? id_node.first : id_edge.first;
    FeatureCommand::Element element = id_node.second ? FeatureCommand::Node : FeatureCommand::Edge;
    m_command = new FeatureCommand(id, feature_cmd.key_, values, mode, element, name, m_verbose);
    m_rule->addCommand(m_command);

    return true;
}

bool CommandMaker::operator()(const grql::ast::modify_command_t &feature_modify)
{
    std::string name = feature_modify.command_;
    /*if(m_verbose)
        std::cerr << "Command " << name << std::endl;*/

    std::pair<unsigned,bool> id_node = m_rule->get_node_id(feature_modify.var_);
    std::pair<unsigned,bool> id_edge = m_rule->get_edge_id(feature_modify.var_);
    if(!id_node.second && !id_edge.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", for " << name << ", ";
        std::cerr << feature_modify.var_ << " does not exist or not refer to a node or an edge" << std::endl;
        return false;
    }

    unsigned id = id_node.second ? id_node.first : id_edge.first;
    ModifyFeatureCommand::Element element = id_node.second ?
                ModifyFeatureCommand::Node : ModifyFeatureCommand::Edge;

    m_command = new ModifyFeatureCommand(id, feature_modify.key1_,
                                         feature_modify.condition_,
                                         feature_modify.replace_,
                                         element, name, m_verbose);
    m_rule->addCommand(m_command);

    return true;
}

bool CommandMaker::operator()(const grql::ast::status_edge_t &status_cmd)
{
    std::string name = status_cmd.command_;

    /*if(m_verbose)
        std::cerr << "Command " << name << std::endl;*/

    std::pair<unsigned, bool> p = m_rule->get_edge_id(status_cmd.var_);

    if(!p.second){
        std::cerr << "In rule " << m_rule->name() << ", for " << name << ", "
                  << "variable " << status_cmd.var_ << " does not exist or not refer to an edge" << std::endl;
        return false;
    }

    m_command = new StatusEdgeCommand(p.first, status_cmd.surfacic_, status_cmd.deep_, name, m_verbose);
    m_rule->addCommand(m_command);
    return true;

    //if(linggraph::checkType(status_cmd.status_)){
    //    m_command = new StatusEdgeCommand(p.first, linggraph::string2type(status_cmd.status_),
    //                                      name, m_verbose);
    //    m_rule->addCommand(m_command);
    //    return true;
    //}
    //else{
    //    std::cerr << "In rule " << m_rule->name() << ", for " << name
    //              << ", unknown edge type: " << status_cmd.status_ << std::endl;
    //    return false;
    //}
}


////////////////////////////////////
// Constraint maker
////////////////////////////////////

std::map<std::string, std::string> ConstraintMaker::features( const std::string & status)
{
    namespace balgo = boost::algorithm;
    using namespace std;

    map<string, string> features;

    vector<string> result;
    balgo::split(result, status, balgo::is_any_of("|"));

    for(vector<string>::iterator it = result.begin(); it != result.end(); ++it){
        vector<string> parts;
        balgo::split(parts, *it, balgo::is_any_of(":"));

        if(parts.size() != 2 )
            continue;
        else
            features.insert( make_pair(parts[0], parts[1]) );
    }

    return features;
}

bool ConstraintMaker::operator ()(const grql::ast::simple_constraint_t &simple)
{
    std::set<std::string> labels( simple.values_.begin(), simple.values_.end() );
    std::copy(simple.values_.begin(), simple.values_.end(), std::inserter(labels, labels.end()));

    std::string status = simple.status_ ? *simple.status_ : "";
    std::map<std::string, std::string> features = this->features(status);

    if(simple.name_ == "redirect_down")
    {
        m_constraint = new RedirectDown(labels, features, m_verbose);
    }
    else if(simple.name_ == "share_down")
    {
        m_constraint = new ShareDown(labels, features, m_verbose);
    }
    else if(simple.name_ == "share_up")
    {
        m_constraint = new ShareUp(labels, features, m_verbose);
    }
    else
    {
        //Error handler to raise an error.
        std::cerr << "In rule " << m_rule->name()
                  << ", unknown constraint " << simple.name_ << std::endl;
        return false;
    }

    VarMaker vm(m_rule);
    if(!boost::apply_visitor(vm, simple.var_))
        return false;

    m_rule->addConstraint(vm.id(), m_constraint);
    return true;

}

bool ConstraintMaker::operator ()(const grql::ast::move_constraint_t& move)
{
    DFAMaker dfam;
    dfam.process_dfa(move.dfa_);

    std::string status = "";
    if(move.status_)
        status = *move.status_;

    if(move.name_ == "move_down"){
        m_constraint = new MoveDown(dfam.dfa(), status, m_verbose);
    }
    else if(move.name_ == "move_share_down")
    {
        m_constraint = new MoveShareDown(dfam.dfa(), status, m_verbose);
    }
    else
    {
        //Error handler to raise an error.
        std::cerr << "In rule " << m_rule->name()
                  << ", unknown constraint " << move.name_ << std::endl;
        return false;
    }

    VarMaker vm(m_rule);
    if(!boost::apply_visitor(vm, move.var_))
        return false;

    m_rule->addConstraint(vm.id(), m_constraint);

    return true;
}

bool ConstraintMaker::operator()(const grql::ast::feature_constraint_t &relabel)
{
    std::set<std::string> labels( relabel.labels_.begin(), relabel.labels_.end() );
    std::copy(relabel.labels_.begin(), relabel.labels_.end(), std::inserter(labels, labels.end()));

    std::string status = relabel.status_ ? *relabel.status_ : "";
    std::map<std::string, std::string> features = this->features(status);

    if(relabel.name_ == "add_feature")
    {
        m_constraint = new AddFeature(labels, relabel.surfacic_label_,
                                      relabel.deep_label_, features, m_verbose);
    }
    else
    {
        std::cerr << "In rule " << m_rule->name()
                  << ", unknown constraint " << relabel.name_ << std::endl;
        return false;
    }

    VarMaker vm(m_rule);
    if(!boost::apply_visitor(vm, relabel.var_))
        return false;

    m_rule->addConstraint(vm.id(), m_constraint);
    return true;
}


/////////////////////////////////
// Var maker
/////////////////////////////////

bool ConstraintMaker::VarMaker::operator ()(const grql::ast::cvar_edge_t &var_nodes)
{
    std::pair<unsigned,bool> src = m_rule->get_node_id(var_nodes.var_source_),
            tar = m_rule->get_node_id(var_nodes.var_target_);

    if(!src.second || !tar.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", ";
        std::cerr << var_nodes.var_source_ << " and/or " << var_nodes.var_target_
                  << " do not exist or not refer to proper nodes" << std::endl;
        return false;
    }

    m_eid = make_pair(src.first, tar.first);
    return true;
}

bool ConstraintMaker::VarMaker::operator ()(const std::string &var_edge)
{
    std::pair<unsigned, bool> id = m_rule->get_edge_id(var_edge);
    if(!id.second)
    {
        std::cerr << "In rule " << m_rule->name() << ", ";
        std::cerr << var_edge << " does not exist or not refer to an edge" << std::endl;
        return false;
    }
    m_eid = id.first;
    return true;
}

}
