#include "ogrecomponents/nac.h"
#include <linguisticgraph/matcher.h>

template<typename T1, typename T2>
struct MultimapKeyFunctor
{
    typedef std::multimap <T1, T2> multi_type;

    T1 operator () (typename multi_type::value_type const & v)
    const
    {
        return v.first;
    }
};

NAC::NAC(){}

ostream &operator<<(ostream &stream, NAC &nac) {
    nac.output(stream);
    return stream;
}

FeatureNAC::FeatureNAC(unsigned id, Mode mode)
    :m_id(id), m_mode(mode){}

ostream& FeatureNAC::output(ostream &stream)
{
    stream << "ID: " << m_id << " on " << (m_mode == FeatureNAC::Node ? "node" : "edge") << std::endl;

    std::multimap<std::string, std::string>::const_iterator it = m_features.begin();
    for(; it != m_features.end(); ++it)
        stream << it->first << " = " << it->second << std::endl;

    return stream;
}

void FeatureNAC::addFeature(std::string key, std::string value)
{
    m_features.insert(std::make_pair(key, value));
}

bool FeatureNAC::keep(LinguisticGraph *sentence,
                               const std::map<unsigned, unsigned> &nodes,
                               const std::map<unsigned, unsigned> &edges)
{
    if(m_mode == Node){
        std::map<unsigned, unsigned>::const_iterator it = nodes.find(m_id);
        if(it != nodes.end()){
            unsigned sent_id = it->second;
            const LinguisticNodeProperties &prop = sentence->nodeProperties(sent_id);
            return test(prop);
        }else{
            return true;
        }
    }else{
        std::map<unsigned, unsigned>::const_iterator it = edges.find(m_id);
        if(it != edges.end()){
            unsigned sent_id = it->second;
            const LinguisticEdgeProperties &prop = sentence->edgeProperties(sent_id);
            return test(prop);
        }else{
            return true;
        }
    }
}

bool FeatureNAC::test(const LinguisticEdgeProperties &edge)
{
    typedef std::multimap<std::string, std::string> map_type;
    map_type edge_map;

    std::set<std::string> keys;
    std::transform(m_features.begin(), m_features.end(),
                   std::inserter(keys, keys.begin()),
                   MultimapKeyFunctor<std::string, std::string>());

    std::set<std::string>::const_iterator it = keys.begin(),
            it_end = keys.end();

    for(; it != it_end; ++it){
        LinguisticEdgeProperties::const_iter_set fit, fit_end;
        const std::string &key = *it;

        if(key == "label"){
            boost::tie(fit, fit_end) = edge.relsCanonical();
            for(; fit != fit_end; ++fit)
                edge_map.insert(make_pair(key, ":"+*fit));

            boost::tie(fit, fit_end) = edge.relsFinal();
            for(; fit != fit_end; ++fit)
                edge_map.insert(make_pair(key, *fit));
        }
    }

    /*std::cerr << "EDGE_MAP content: " << std::endl;
    for(map_type::const_iterator it = edge_map.begin(); it != edge_map.end(); ++it)
        std::cerr << it->first << " = " << it->second << std::endl;*/

    map_type result;
    std::set<map_type::value_type> s1(m_features.begin(), m_features.end());
    std::set<map_type::value_type> s2(edge_map.begin(), edge_map.end());
    std::set_intersection(s1.begin(), s1.end(),
                          s2.begin(), s2.end(),
                          std::inserter(result, result.end()));

    if(result.size() == 0)
        return true; //Ok the matching is fine
    else
        return false; //No the matching need to be skip
}

bool FeatureNAC::test(const LinguisticNodeProperties &node)
{
    std::set<std::string> keys;
    std::transform(m_features.begin(), m_features.end(),
                   std::inserter(keys, keys.begin()),
                   MultimapKeyFunctor<std::string, std::string>());

    typedef std::multimap<std::string, std::string> map_type;
    map_type node_map;

    std::set<std::string>::const_iterator it = keys.begin(),
            it_end = keys.end();
    for(; it != it_end; ++it){
        LinguisticNodeProperties::const_feature_iter fit, fit_end;
        const std::string &key = *it;

        if(key == "token"){
            boost::tie(fit, fit_end) = node.tokens();
        }else if(key == "lemma"){
            boost::tie(fit, fit_end) = node.lemmas();
        }else if(key == "hypercat"){
            boost::tie(fit, fit_end) = node.hyperposs();
        }else if(key == "cat"){
            boost::tie(fit, fit_end) = node.poss();
        }else{
            boost::tie(fit, fit_end) = node.features(key);
        }
        std::copy(fit, fit_end, std::inserter(node_map, node_map.begin()));
    }

    /*std::cerr << "NODE_MAP content: " << std::endl;
    for(map_type::const_iterator it = node_map.begin(); it != node_map.end(); ++it)
        std::cerr << it->first << " = " << it->second << std::endl;*/

    map_type result;
    std::set<map_type::value_type> s1(m_features.begin(), m_features.end());
    std::set<map_type::value_type> s2(node_map.begin(), node_map.end());
    std::set_intersection(s1.begin(), s1.end(),
                          s2.begin(), s2.end(),
                          std::inserter(result, result.end()));

    if(result.size() == 0)
        return true; //Ok the matching is fine
    else
        return false; //No the matching need to be skip
}


/////////////////////////////////////////////////////////
////// NAC on subgraph
/////////////////////////////////////////////////////////

GraphNAC::GraphNAC(LinguisticGraph *graph, SetPair ids)
    :m_graph(graph), m_ids(ids)
{

}

GraphNAC::~GraphNAC()
{
    delete m_graph;
}

ostream& GraphNAC::output(ostream &stream)
{
    LinguisticGraph::node_iter nit, nit_end;
    for(boost::tie(nit, nit_end) = m_graph->nodes(); nit != nit_end; ++nit){
        const LinguisticNodeProperties &n = m_graph->nodeProperties(*nit);

        if( m_graph->nodeOutDegree(*nit) > 0 ){
            stream << n << " -(";

            LinguisticGraph::out_edge_iter oit, oit_end;
            for(boost::tie(oit, oit_end) = m_graph->outEdges(n.id); oit != oit_end; ++oit){
                stream << m_graph->edgeProperties(*oit)
                       << ")-> " << m_graph->nodeProperties( m_graph->targetNode(*oit) )
                       << std::endl;
            }
        }
    }
    return stream;
}

// If we find the joint id in the nac matching :
// 1. This means that the rule matched.
// 2. We need to determine if it matched the right node
//  --> We check the id in sentence matched by the NAC
//  --> We check the id in sentence matched by the rule
//      --> If it the same, we discard the matching for the rule

// Let's take an example to understand the problem:
// _John eats an apple and drinks an orange juice_

// We want to make sure that "drinks" does not have a subject
// if it's the case, it's a subject ellipsis and we can annotate it.
// if it's not, it's just a basic coordination, we don't need to do anything

// We have a match saying :
// "v -coord-> and -dep.coord-> v2"

// We have a NAC rule saying :
// "v2 -suj-> []"

//The NAC will match once :
// 1. John eats
// BUT "eats" it's not denoted by v2
// that's exactly what we check in this method

// On the contrary, for instance : _John eats an apple and Mary drinks an orange juice_
// The NAC will match twice :
// 1. John eats
// 2. Mary drinks
// AND "drinks" is denoted by v2
// We need to discard the match
bool GraphNAC::keep(LinguisticGraph *sentence,
          const std::map<unsigned, unsigned> & nodes,
          const std::map<unsigned, unsigned> & edges)
{
    typedef vector<map<unsigned, unsigned> > MatchesType;
    typedef std::pair<MatchesType, MatchesType> MatchesPairType;

    GraphsMatcher m;
    MatchesPairType matching = m(m_graph, sentence);

    if(matching.first.size() == 0){
        //std::cerr << "No matching for the NAC Graph" << std::endl;
        return true;
    }else{
        MatchesType nac_nodes_matches = matching.first;

        //std::cerr << nac_nodes_matches.size() << " matching for the NAC Graph" << std::endl;

        for(int i = 0; i < nac_nodes_matches.size(); i++){
            std::map<unsigned, unsigned> nac_nodes = nac_nodes_matches[i];            

            /*for(std::map<unsigned, unsigned>::const_iterator it = nac_nodes.begin(); it != nac_nodes.end(); ++it)
                std::cerr << it->first << " -> " << it->second << std::endl;*/

            set<unsigned> nac_nids = m_ids.first;
            for(set<unsigned>::const_iterator it_ids = nac_nids.begin();
                it_ids != nac_nids.end(); ++it_ids)
            {
                unsigned nac_node_id = *it_ids;
                if(nac_nodes.find(nac_node_id) != nac_nodes.end())
                {
                    unsigned id_sentence_nac  = nac_nodes.find(nac_node_id)->second;
                    unsigned id_sentence_rule = nodes.find(nac_node_id)->second;
                    if(id_sentence_nac == id_sentence_rule)
                        return false;
                }
            }
        }

        MatchesType nac_edges_matches = matching.second;
        for(int i = 0; i < nac_edges_matches.size(); i++){
            std::map<unsigned, unsigned> nac_edges = nac_edges_matches[i];

            set<unsigned> nac_nids = m_ids.second;
            for(set<unsigned>::const_iterator it_ids = nac_nids.begin();
                it_ids != nac_nids.end(); ++it_ids)
            {
                unsigned nac_edge_id = *it_ids;
                if(nac_edges.find(nac_edge_id) != nac_edges.end())
                {
                    unsigned id_sentence_nac  = nac_edges.find(nac_edge_id)->second;
                    unsigned id_sentence_rule = edges.find(nac_edge_id)->second;
                    if(id_sentence_nac == id_sentence_rule)
                        return false;
                }
            }
        }

        return true;
    }
}

/////////////////////////////////////////////////////////
////// Precedence NAC
/////////////////////////////////////////////////////////
PrecedenceNAC::PrecedenceNAC(std::vector<unsigned> ids)
    :m_ids(ids)
{}

PrecedenceNAC::~PrecedenceNAC()
{}

ostream& PrecedenceNAC::output(ostream &stream)
{
    for(int i = 0; i < m_ids.size(); i++){
        if(i)
            stream << "< " << m_ids[i];
        else
            stream << m_ids[i];
    }

    return stream;
}

//Precedence works as follow =
//Nodes in graph are in ascending order (representing linear order in the sentence)
//We impose precedence of certain nodes
//So our vector is holding precedence.
//For example : it can be :
// 3 < 1 < 2
// which, for instance, corresponds in graph to 1 < 2 < 6
// the nodes map gives us the correspondance (3, 1)(1, 2)(2, 6).
// To test the precedence, we iterate over the vector
// and test if the last id is less than the subsequent one
// if it's not the case, we discard the match, because
// the precedence is not kept.
// Let's imagine the following example:
// _John <-suj- loves -obj-> Mary_
//
// We have a match like this :
// x[lemma:"love"] -[label:obj]-> y[lemma:"Mary"]
//
// We impose that x < y
// So that it to say love < Mary
// We know for sure that the graph sentence will have id as follow :
// John = 0 / loves = 1 / Mary = 2
// Lets say that x = 6 and y = 5
// So the nodes map will be the following :
// nodes(6, 1)(5, 2)
// When we iterate over the vector<6, 5> and we translate
// the id into the sentence id we will have
// current_node_sentence = 1
// nodes.find(m_ids[i])->second = 2
// and 1 < 2 --> So we're good.
// precedence is ok.
bool PrecedenceNAC::keep(LinguisticGraph *sentence,
          const std::map<unsigned, unsigned> & nodes,
          const std::map<unsigned, unsigned> & /*edges*/)
{
    if(m_ids.size() == 0 )
        return true;

    unsigned current_node_sentence = nodes.find(m_ids[0])->second;

    for(int i = 1; i < m_ids.size(); i++){
        if(current_node_sentence > nodes.find(m_ids[i])->second )
            return false;
    }

    return true;
}
