#include "ogrecomponents/rewriter.h"
#include <linguisticgraph/matcher.h>

#include <boost/lexical_cast.hpp>

namespace ogrecomponents{

// WARNING / TODO / WARNING
// If user specifies (x,y) for example a constraint_edge will be created
// If he specifies "e" a constraint will be made on the edge named "e"
// We need to check if there exists an edge between (x,y) which satisfies
// the constraint properties so that we do not create an artifical edge
// if we do not want to.
struct ConstraintIdVisitor: boost::static_visitor<unsigned>
{
    ConstraintIdVisitor(LinguisticGraph *sentence,
                        const std::map<unsigned, unsigned> &nodes,
                        const std::map<unsigned, unsigned> &edges)
        :m_sentence(sentence), m_nodes(nodes), m_edges(edges){}

    unsigned operator()(const unsigned eid)
    {
        return m_edges.find(eid)->second;
    }

    unsigned operator()(const std::pair<unsigned, unsigned> ids)
    {
        unsigned src = m_nodes.find(ids.first)->second,
                 tar = m_nodes.find(ids.second)->second;

        unsigned id = m_sentence->lastEdgeId()+1;

        LinguisticEdgeProperties edge(id,
                                      "__ogre_constraint__",boost::lexical_cast<std::string>(id));
        m_sentence->addEdge(src, tar, edge);
        return edge.id;
    }

    LinguisticGraph *m_sentence;
    const std::map<unsigned, unsigned> &m_nodes, &m_edges;
};

bool Rewriter::rewrite(LinguisticGraph *sentence,
                       LinguisticGraph *result, Compiler *compiler)
{
    ConstraintVector constraints;
    bool is_rewritten = false;
    Compiler::RulesConstIter it, eit;

    for(boost::tie(it, eit) = compiler->rules(); it != eit; ++it)
    {
        Rule *r = *it;
        if(m_verbose >= DEBUG)
            std::cerr << "Applying rule " << r->name() << std::endl;

        MatchesPairType matching = match(r, sentence);
        MatchesType& match_nodes = matching.first, match_edges = matching.second;        

        if(match_nodes.size() == 0)
            continue;

        if(m_onlymatch){

        }

        if(firstStageRewriting(result, r, match_nodes, match_edges))
            is_rewritten = true;
        makeConstraint(result, r, match_nodes, match_edges, constraints);

        if(m_verbose)
            std::cerr << "-------------------------------------------------------" << std::endl;
    }


    //Second stage
    if(m_verbose >= DEBUG && constraints.size() > 0)
        std::cerr << "Applying " << constraints.size() <<  " constraints for the sentence" << std::endl;
    bool constraints_applied = secondStageRewriting(result, constraints);

    return (constraints_applied || is_rewritten);
}

Rewriter::MatchesPairType Rewriter::match(Rule *rule,
                     LinguisticGraph *sentence)
{
    GraphsMatcher m;
    MatchesPairType matching = m(rule->graph(), sentence);
    MatchesType& match_nodes = matching.first, match_edges = matching.second;

    if(m_verbose >= DEBUG && match_nodes.size() > 0){
        std::cerr << match_nodes.size() << " matching(s) before NAC application" << std::endl;
    }

    //Test Negative Application Conditions
    if(rule->NACCount() > 0)
        matching = testNACs(rule, sentence, match_nodes, match_edges);

    if(m_verbose >= DEBUG)
        std::cerr << match_nodes.size() << " matching(s) for that sentence" << std::endl;

    return matching;
}


Rewriter::MatchesPairType Rewriter::testNACs(Rule *rule,
                        LinguisticGraph *sentence,
                        vector<map<unsigned, unsigned> >&nodes,
                        vector<map<unsigned, unsigned> >&edges)
{
    MatchesType new_nodes, new_edges;
    for(int i = 0; i < nodes.size(); i++)
    {
        map<unsigned, unsigned> map_nodes = nodes[i];
        map<unsigned, unsigned> map_edges;
        if(nodes.size() == edges.size())
            map_edges = edges[i];

        Rule::NacVectorConstIter nac_it, nac_end;
        bool keep_matching = false;
        for(boost::tie(nac_it, nac_end) = rule->NACs(); nac_it != nac_end; ++nac_it)
        {
            NAC *nac = *nac_it;
            //std::cerr << "NAC: " << *nac << std::endl;

            if(!nac->keep(sentence, map_nodes, map_edges)){
                keep_matching = false;
                break;
            }else{
                keep_matching = true;
            }
        }

        if( keep_matching )
        {
            new_nodes.push_back(map_nodes);
            if(nodes.size() == edges.size())
                new_edges.push_back(map_edges);
            //std::cerr << "The matching is fine" << std::endl;
        }
    }

    return std::make_pair(new_nodes, new_edges);
}


bool Rewriter::firstStageRewriting(LinguisticGraph *result,
                                   Rule *rule,
                                   const vector<map<unsigned, unsigned> > &matching_nodes,
                                   const vector<map<unsigned, unsigned> > &matching_edges)
{
    bool local_rewriting = false;

    for(int i = 0; i < matching_nodes.size(); i++)
    {
        map<unsigned, unsigned> m_nodes = matching_nodes[i];
        if(m_nodes.size() > 0)
        {
            local_rewriting = true;

            if(m_verbose >= DEBUG)
            {
                std::cerr << "matching nodes: " << std::endl;
                for(map<unsigned, unsigned>::iterator it = m_nodes.begin(); it != m_nodes.end(); ++it)
                {
                    try{
                        std::cerr << "\t" << rule->graph()->nodeProperties(it->first) << " -> " << result->nodeProperties(it->second) << std::endl;
                    }catch(std::out_of_range&){
                        std::cerr << "node deleted during rewriting..." << std::endl;
                    }
                }

                //std::cerr << "Number of commands to apply: " << rule->commandsCount() << std::endl;
            }

            map<unsigned, unsigned> m_edges;
            if(matching_edges.size() == matching_nodes.size())
                m_edges = matching_edges[i];

            if(m_verbose >= DEBUG)
            {
                std::cerr << "matching edges: " << std::endl;
                for(map<unsigned, unsigned>::iterator it = m_edges.begin(); it != m_edges.end(); ++it)
                {
                    try{
                        std::cerr << "\t" << rule->graph()->edgeProperties(it->first) << " -> " << result->edgeProperties(it->second) << std::endl;
                        std::cerr << "\t" << result->nodeProperties( result->sourceNode(it->second) )
                                  << result->nodeProperties( result->targetNode(it->second) ) << std::endl;
                    }catch(std::out_of_range&){
                        std::cerr << "edge deleted during rewriting..." << std::endl;
                    }
                }

                std::cerr << "Number of commands to apply: " << rule->commandsCount() << std::endl;
            }

            Rule::CommandVectorConstIter cit, ecit;
            for(boost::tie(cit, ecit) = rule->commands(); cit != ecit; ++cit)
            {
                Command *command = *cit;
                command->process(m_nodes, m_edges, result);
            }
        }
        else
            local_rewriting = false;
    }

    return local_rewriting;
}

void Rewriter::makeConstraint(LinguisticGraph *result,
                              Rule *rule,
                              const vector<map<unsigned, unsigned> > &nodes,
                              const vector<map<unsigned, unsigned> > &edges,
                              ConstraintVector &constraints)
{
    if(rule->edgesWithConstraintCount() == 0)
        return;

    if(nodes.size() != edges.size())
    {
        std::cerr << "Matchings between nodes and edges are different..." << std::endl;
        std::cerr << "Unable to attach constraint(s)..." << std::endl;
        return;
    }

    Rule::ConstraintVectorConstIter cons_it, cons_endit;
    for(boost::tie(cons_it, cons_endit) = rule->constraints(); cons_it != cons_endit; ++cons_it)
    {
        std::pair<Rule::ConstraintIdType, Constraint*> p = *cons_it;

        for(int i = 0; i < nodes.size(); i++)
        {
            ConstraintIdVisitor cv(result, nodes[i], edges[i]);
            unsigned id = boost::apply_visitor(cv, p.first);
            Constraint* c = p.second;
            c->resetHistoric();

            constraints.push_back(make_pair(id, c));
        }
    }
}

bool Rewriter::secondStageRewriting(LinguisticGraph *result, const ConstraintVector& constraints)
{
    bool constraints_applied = false;

    typedef ConstraintVector::const_iterator ConsIterator;

    std::set<unsigned> constraint_edges;

    //Constraints rewriting
    bool redo = true;
    while(redo)
    {
        redo = false;

        for(ConsIterator it = constraints.begin(); it != constraints.end(); ++it)
        {
            std::pair<unsigned, Constraint*> p = *it;

            unsigned id = p.first;
            Constraint* c = p.second;

            if(result->edgeProperties(id).relFinal() == "__ogre_constraint__")
                constraint_edges.insert(id);

            if(!redo)
                redo = c->trigger(result, id);
            else
                c->trigger(result, id);
        }

        constraints_applied = redo;
    }

    std::set<unsigned>::const_iterator it = constraint_edges.begin();
    for(; it != constraint_edges.end(); ++it)
    {
        if(result->existsEdge(*it))
            result->removeEdge(*it);
    }
    return constraints_applied;
}

}
