#include <boost/test/unit_test.hpp>
#include <boost/assign/list_of.hpp>
#include <iostream>

#include "ogrecomponents/command.h"

using namespace ogrecomponents;

struct GraphFixture
{
    GraphFixture()
    {
        graph = new LinguisticGraph;
        LinguisticNodeProperties node1(0, "Jean", "Jean", "N", "NPP"),
                node2(1, "aime", "aimer", "V", "V"),
                node3(2, "manger", "manger", "V", "VINF"),
                node4(3, "des", "le", "D", "DET"),
                node5(4, "pommes", "pomme", "N", "NC");

        LinguisticEdgeProperties edge1(0, "suj", "suj"),
                edge2(1, "obj", "obj"),
                edge3(2, "det", "det"),
                edge4(3, "obj", "obj");

        graph->addNode(node1.id, node1);
        graph->addNode(node2.id, node2);
        graph->addNode(node3.id, node3);
        graph->addNode(node4.id, node4);
        graph->addNode(node5.id, node5);

        // aime -suj-> Jean
        graph->addEdge(1, 0, edge1);
        // aime -obj-> manger
        graph->addEdge(1, 2, edge2);
        // manger -obj-> pommes
        graph->addEdge(2, 4, edge4);
        // pommes -det-> des
        graph->addEdge(4, 3, edge3);


        //Dummy rule graph
        rule = new LinguisticGraph;
        LinguisticNodeProperties rnode1(5, "Jean", "Jean", "N", "NPP"),
                rnode2(6, "aime", "aimer", "V", "V");
        LinguisticEdgeProperties redge1(5, "suj", "suj");

        rule->addNode(rnode1.id, rnode1);
        rule->addNode(rnode2.id, rnode2);
        // aime -suj-> Jean
        rule->addEdge(6, 5, redge1);

    }

    LinguisticGraph *graph;
    LinguisticGraph *rule;
};

BOOST_FIXTURE_TEST_SUITE(CommandSuite, GraphFixture)

BOOST_AUTO_TEST_CASE(AddEdgeCommandTest){
    unsigned rule_id_src = 5, rule_id_tar = 6;

    AddEdgeCommand *addEdgeCommand = new AddEdgeCommand(rule_id_src, rule_id_tar, "add_edge", NONE);
    addEdgeCommand->add_feature("label", "suj:suj");
    addEdgeCommand->add_feature("deep", "y");
    addEdgeCommand->add_feature("surfacic", "n");

    std::map<unsigned, unsigned> matching_nodes =
            boost::assign::map_list_of(rule_id_tar, 0)(rule_id_src, 2);
    std::map<unsigned, unsigned> matching_edges;

    unsigned old_edge_size = graph->edgesSize();
    addEdgeCommand->process(matching_nodes, matching_edges, graph);

    BOOST_CHECK_EQUAL(graph->edgesSize(), old_edge_size+1);
    BOOST_CHECK_EQUAL(graph->edgeProperties(graph->edgesSize()-1).relFinal(), "suj");
    BOOST_CHECK_EQUAL(graph->edgeProperties(graph->edgesSize()-1).relCanonical(), "suj");
    BOOST_CHECK_EQUAL(graph->edgeProperties(graph->edgesSize()-1).type(), LinguisticEdgeProperties::Deep);
}

BOOST_AUTO_TEST_CASE(AddEdgeCommandTwiceTest){
    unsigned rule_id_src = 5, rule_id_tar = 6;

    AddEdgeCommand *addEdgeCommand = new AddEdgeCommand(rule_id_src, rule_id_tar, "add_edge", NONE);
    addEdgeCommand->add_feature("label", "suj:suj");
    addEdgeCommand->add_feature("surfacic", "n");
    addEdgeCommand->add_feature("deep", "y");

    std::map<unsigned, unsigned> matching_nodes =
            boost::assign::map_list_of(rule_id_tar, 0)(rule_id_src, 2);
    std::map<unsigned, unsigned> matching_edges;

    unsigned old_edge_size = graph->edgesSize();
    addEdgeCommand->process(matching_nodes, matching_edges, graph);

    unsigned eid = graph->edgesSize()-1;

    BOOST_CHECK_EQUAL(graph->edgesSize(), old_edge_size+1);
    BOOST_CHECK_EQUAL(graph->edgeProperties(eid).relFinal(), "suj");
    BOOST_CHECK_EQUAL(graph->edgeProperties(eid).relCanonical(), "suj");
    BOOST_CHECK_EQUAL(graph->edgeProperties(eid).type(), LinguisticEdgeProperties::Deep);
    BOOST_CHECK_EQUAL(graph->nodeProperties( graph->sourceNode(eid) ).id, 2);
    BOOST_CHECK_EQUAL(graph->nodeProperties( graph->targetNode(eid) ).id, 0);

    old_edge_size = graph->edgesSize();
    addEdgeCommand->process(matching_nodes, matching_edges, graph);

    BOOST_CHECK_EQUAL(graph->edgesSize(), old_edge_size);
    BOOST_CHECK_EQUAL(graph->nodeProperties( graph->sourceNode(eid) ).id, 2);
    BOOST_CHECK_EQUAL(graph->nodeProperties( graph->targetNode(eid) ).id, 0);
}


BOOST_AUTO_TEST_CASE(RemoveNodeCommandTest)
{    
    unsigned rule_id = 5;
    RemoveCommand *removeCommand = new RemoveCommand(rule_id, RemoveCommand::Node, "remove_node", NONE);

    //verbe in rule corresponds to aime in graph
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(5, 0),
            matching_edges;

    unsigned old_node_size = graph->nodesSize();
    unsigned old_edge_size = graph->edgesSize();
    removeCommand->process(matching_nodes, matching_edges, graph);

    BOOST_CHECK_EQUAL(graph->nodesSize(), old_node_size - 1);
    BOOST_CHECK_EQUAL(graph->edgesSize(), old_edge_size - 1);
    BOOST_CHECK_THROW(graph->edgeProperties(0), std::out_of_range);
    //BOOST_CHECK_THROW(graph->nodeProperties(0), std::out_of_range);
}

BOOST_AUTO_TEST_CASE(RemoveEdgeCommandTest)
{
    unsigned rule_id = 5;

    RemoveCommand *removeCommand = new RemoveCommand(rule_id, RemoveCommand::Edge, "remove_edge", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(6, 1),
            matching_edges = boost::assign::map_list_of(rule_id, 0);

    unsigned old_edge_size = graph->edgesSize();
    removeCommand->process(matching_nodes, matching_edges, graph);

    BOOST_CHECK_EQUAL(graph->edgesSize(), old_edge_size - 1);
    BOOST_CHECK_THROW(graph->edgeProperties(0), std::out_of_range);
}

BOOST_AUTO_TEST_CASE(AddFeatureCommandOnNodeTest)
{
    unsigned rule_id = 5;

    std::set<std::string> values = boost::assign::list_of("n");
    FeatureCommand *featureCommand = new FeatureCommand(rule_id, "void", values, FeatureCommand::Add, FeatureCommand::Node, "add_feature", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0),
            matching_edges;

    featureCommand->process(matching_nodes, matching_edges, graph);

    const LinguisticNodeProperties& prop = graph->nodeProperties((unsigned)0);
    BOOST_CHECK_EQUAL(prop.feature("void"), "n");
}

BOOST_AUTO_TEST_CASE(AddFeatureCommandOnNodeWithVariableReferenceTest)
{
    unsigned rule_id = 5;
    boost::tuples::tuple<unsigned, std::string, FeatureCommand::Element> values = boost::tuples::make_tuple(6, "token", FeatureCommand::Node);
    FeatureCommand *featureCommand = new FeatureCommand(rule_id, "void", values, FeatureCommand::Add, FeatureCommand::Node, "add_feature", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0)(6, 1),
            matching_edges;

    featureCommand->process(matching_nodes, matching_edges, graph);

    const LinguisticNodeProperties& prop = graph->nodeProperties((unsigned)0);
    BOOST_CHECK_EQUAL(prop.feature("void"), "aime");
}

BOOST_AUTO_TEST_CASE(AddFeatureCommandOnEdgeWithVariableReferenceTest)
{
    unsigned rule_id = 5;
    boost::tuples::tuple<unsigned, std::string, FeatureCommand::Element> values = boost::tuples::make_tuple(6, "token", FeatureCommand::Node);
    FeatureCommand *featureCommand = new FeatureCommand(rule_id, "label", values, FeatureCommand::Add, FeatureCommand::Edge, "add_feature", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0)(6, 1),
            matching_edges = boost::assign::map_list_of(5, 0); //Link between aime and Jean (suj:suj)

    featureCommand->process(matching_nodes, matching_edges, graph);

    const LinguisticEdgeProperties& prop = graph->edgeProperties((unsigned)0);
    BOOST_CHECK_EQUAL(prop.relFinal(), "aime");
}

BOOST_AUTO_TEST_CASE(ModifyFeatureCommandOnNodeTest)
{
    unsigned rule_id = 5;

    ModifyFeatureCommand *featureCommand = new ModifyFeatureCommand(rule_id, "void", "y", "m",
                                                                    ModifyFeatureCommand::Node, "modify_feature", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0),
            matching_edges;

    LinguisticNodeProperties& prop = graph->nodeProperties((unsigned)0);
    prop.addFeature("void", "y");
    BOOST_CHECK_EQUAL(prop.feature("void"), "y");

    featureCommand->process(matching_nodes, matching_edges, graph);

    BOOST_CHECK_EQUAL(prop.feature("void"), "m");
}

BOOST_AUTO_TEST_CASE(ModifyFeatureCommandOnEdgeTest)
{
    unsigned rule_id = 5;

    ModifyFeatureCommand *featureCommand = new ModifyFeatureCommand(rule_id, "label", "suj", "obj",
                                                                    ModifyFeatureCommand::Edge, "modify_feature", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0),
            matching_edges;

    featureCommand->process(matching_nodes, matching_edges, graph);
    LinguisticEdgeProperties& prop = graph->edgeProperties((unsigned)0);
    BOOST_CHECK_EQUAL(prop.relFinal(), "obj");
    BOOST_CHECK_EQUAL(prop.relCanonical(), "suj");

    ModifyFeatureCommand *featureCommandOnCanonicalLabel = new ModifyFeatureCommand(rule_id, "label", ":suj", ":obj",
                                                                    ModifyFeatureCommand::Edge, "modify_feature", NONE);
    featureCommandOnCanonicalLabel->process(matching_nodes, matching_edges, graph);
    BOOST_CHECK_EQUAL(prop.relFinal(), "obj");
    BOOST_CHECK_EQUAL(prop.relCanonical(), "obj");
}

BOOST_AUTO_TEST_CASE(AppendFeatureCommandOnNodeTest)
{
    unsigned rule_id = 5;

    std::set<std::string> values = boost::assign::list_of("n");
    FeatureCommand *featureCommand =
            new FeatureCommand(rule_id, "void", values,
                               FeatureCommand::Append,
                               FeatureCommand::Node, "add_feature", NONE);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0),
            matching_edges;

    LinguisticNodeProperties& prop = graph->nodeProperties((unsigned)0);
    prop.addFeature("void", "y");
    featureCommand->process(matching_nodes, matching_edges, graph);

    std::set<std::string> given, expected = boost::assign::list_of("n")("y");

    LinguisticNodeProperties::const_feature_iter it, eit;
    for(boost::tie(it, eit) = prop.features("void"); it != eit; ++it)
        given.insert(it->second);

    BOOST_CHECK_EQUAL_COLLECTIONS(given.begin(), given.end(), expected.begin(), expected.end());
}

BOOST_AUTO_TEST_CASE(RemoveFeatureCommandOnNodeTest)
{
    /*unsigned rule_id = 5;

    std::set<std::string> values = boost::assign::list_of("y");
    FeatureCommand *featureCommand = new FeatureCommand(rule_id, "void", values, FeatureCommand::Remove, FeatureCommand::Node, "modify_feature", false);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(rule_id, 0),
            matching_edges;

    LinguisticNodeProperties& prop = graph->nodeProperties((unsigned)0);
    prop.addFeature("void", "y");
    BOOST_CHECK_EQUAL(prop.feature("void"), "y");

    featureCommand->process(matching_nodes, matching_edges, graph);

    BOOST_CHECK_EQUAL(prop.feature("void"), "");*/
}

BOOST_AUTO_TEST_CASE(CloneEdgeFeatureCommandNoChangeTest)
{
    unsigned rule_id = 5;
    MoveEdgeCommand *move = new MoveEdgeCommand(rule_id, 7, 5,
                                                MoveEdgeCommand::Clone,
                                                "clone_edge", DEBUG);
    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(7, 2)(5, 0),
            matching_edges = boost::assign::map_list_of(5, 0);

    unsigned edges_size = graph->edgesSize();

    move->process(matching_nodes, matching_edges, graph);

    const LinguisticEdgeProperties &old = graph->edgeProperties(unsigned(0)),
            &clone = graph->edgeProperties(graph->lastEdgeId());

    BOOST_CHECK_EQUAL(edges_size+1, graph->edgesSize());
    BOOST_CHECK_EQUAL(old.compare(clone), true);
}

BOOST_AUTO_TEST_CASE(CloneEdgeFeatureCommandChangeTest)
{
    unsigned rule_id = 5;
    MoveEdgeCommand *move = new MoveEdgeCommand(rule_id, 7, 5,
                                                MoveEdgeCommand::Clone,
                                                "clone_edge", DEBUG);
    move->add_feature("deep", "n");
    move->add_feature("label", "suj:_");

    std::map<unsigned, unsigned> matching_nodes = boost::assign::map_list_of(7, 2)(5, 0),
            matching_edges = boost::assign::map_list_of(5, 0);

    unsigned edges_size = graph->edgesSize();

    move->process(matching_nodes, matching_edges, graph);

    const LinguisticEdgeProperties &old = graph->edgeProperties(unsigned(0)),
            &clone = graph->edgeProperties(graph->lastEdgeId());

    BOOST_CHECK_EQUAL(edges_size+1, graph->edgesSize());
    BOOST_CHECK_EQUAL(old.compare(clone), false);
    BOOST_CHECK_EQUAL(clone.typeString(), "S");
    BOOST_CHECK_EQUAL(clone.relCanonical(), "_");
    BOOST_CHECK_EQUAL(clone.relFinal(), "suj");
}

BOOST_AUTO_TEST_SUITE_END()
