#include <boost/test/unit_test.hpp>
#include <boost/assign/list_of.hpp>
#include <iostream>

#include <boost/shared_ptr.hpp>

#include "ogrecomponents/compiler.h"

#include <linguisticgraph/matcher.h>

namespace oc = ogrecomponents;
namespace ba = boost::assign;

class CompilerFixture
{
public:

    CompilerFixture()
    {
        BOOST_TEST_MESSAGE("Compiler Fixture setup");

        grql_ellipsis =
                    "rule ellipsis{\n"
                        "match{\n"
                            "v[hypercat:\"V\"] (\n"
                                "-[label:\"suj\"]-> suj,\n"
                                "-[label:\"coord\"]-> [hypercat:\"C\"] ( -[label:\"dep.coord\"]-> v2[hypercat:\"V\"] )\n"
                            ")\n"
                        "}\n"
                    "}";

        grql_ellipsis_command =
                "rule ellipsis{\n"
                    "match{\n"
                        "v[hypercat:\"V\"] (\n"
                            "-[label:\"suj\"]-> suj,\n"
                            "-[label:\"coord\"]-> [hypercat:\"C\"] ( -[label:\"dep.coord\"]-> v2[hypercat:\"V\"] )\n"
                        ")\n"
                    "}\n"
                    "commands{\n"
                        "add_feature(v, \"void\", \"y\")\n"
                        "add_feature(v, \"test\", \"n\")\n"
                    "}\n"
                "}";

        grql_ellipsis_constraint =
                "rule ellipsis{\n"
                    "match{\n"
                        "v[hypercat:\"V\"] (\n"
                            "-[label:\"suj\"]-> suj,\n"
                            "-[label:\"coord\"]-> [hypercat:\"C\", lemma:\"et\"|\"ou\"|\"mais\"|\"ni\"] -[label:\"dep.coord\"]-> v2[hypercat:\"V\"]\n"
                        ")\n"
                    "}\n"
                    "constraints{\n"
                        "share_up( (v2, v) , \"suj\", \"surfacic:n\" )\n"
                    "}\n"
                "}\n";

        grql_status_edge_command =
                "rule status{\n"
                    "match{\n"
                        "x -e[label:\"aux.tps\"]-> v\n"
                    "}\n"
                    "commands{\n"
                        "edge_status(e, true, false)\n"
                    "}\n"
                "}";

        grql_nac_rule =
                "rule nac{\n"
                    "match{\n"
                        "x -e[label:\"aux.tps\"]-> v\n"
                    "}\n"
                    "negative{\n"
                        "x -[label:\"suj\"]-> []"
                    "}\n"
                    "commands{\n"
                        "edge_status(e, true, false)\n"
                    "}\n"
                "}";

        grql_reference_rule =
                "rule ref{\n"
                    "match{\n"
                        "x -e[label:'aux.tps']-> v\n"
                    "}\n"
                    "commands{\n"
                        "add_feature(x, 'token', v['cat'])"
                    "}\n"
                "}";

        ellipsis_graph = new LinguisticGraph;
        LinguisticNodeProperties jean(0, "Jean", "Jean", "N", "NPP"),
                mange(1, "mange", "manger", "V", "V"),
                et(2, "et", "et", "C", "CC"),
                dort(3, "dort", "dormir", "V", "V");

        LinguisticEdgeProperties suj(0, "suj", ""),
                coord(1, "coord", ""),
                depcoord(2, "dep.coord", "");
        suj.setOriginal(true);
        coord.setOriginal(true);
        depcoord.setOriginal(true);

        ellipsis_graph->addNode(jean.id, jean);
        ellipsis_graph->addNode(mange.id, mange);
        ellipsis_graph->addNode(et.id, et);
        ellipsis_graph->addNode(dort.id, dort);

        ellipsis_graph->addEdge(mange.id, jean.id, suj);
        ellipsis_graph->addEdge(mange.id, et.id, coord);
        ellipsis_graph->addEdge(et.id, dort.id, depcoord);

        //Graphs to test the NAC
        test_nac_graph_discard = new LinguisticGraph;
        test_nac_graph_keep    = new LinguisticGraph;

        LinguisticNodeProperties
                nac_jean(1, "Jean", "Jean", "N", "NPP"),
                nac_a(2, "a", "avoir", "V", "V"),
                nac_aime(3, "aimé", "aimer", "V", "VPP"),
                nac_marie(4, "Marie", "Marie", "N", "NPP");

        LinguisticEdgeProperties
                nac_suj(1, "suj", ""),
                nac_auxtps(2, "aux.tps", ""),
                nac_obj(3, "obj");

        test_nac_graph_keep->addNode(nac_a.id, nac_a);
        test_nac_graph_keep->addNode(nac_aime.id, nac_aime);
        test_nac_graph_keep->addNode(nac_marie.id, nac_marie);

        test_nac_graph_keep->addEdge(nac_aime.id, nac_a.id, nac_auxtps);
        test_nac_graph_keep->addEdge(nac_aime.id, nac_marie.id, nac_obj);

        ///////
        test_nac_graph_discard->addNode(nac_jean.id, nac_jean);
        test_nac_graph_discard->addNode(nac_a.id, nac_a);
        test_nac_graph_discard->addNode(nac_aime.id, nac_aime);
        test_nac_graph_discard->addNode(nac_marie.id, nac_marie);

        test_nac_graph_discard->addEdge(nac_aime.id, nac_jean.id, nac_suj);
        test_nac_graph_discard->addEdge(nac_aime.id, nac_a.id, nac_auxtps);
        test_nac_graph_discard->addEdge(nac_aime.id, nac_marie.id, nac_obj);
    }

    ~CompilerFixture()
    {
        BOOST_TEST_MESSAGE("Compiler Fixture teardown");
        delete ellipsis_graph;
        delete test_nac_graph_discard;
        delete test_nac_graph_keep;
    }

    std::string grql_ellipsis;
    std::string grql_ellipsis_command;
    std::string grql_ellipsis_constraint;
    std::string grql_status_edge_command;
    std::string grql_nac_rule;
    std::string grql_reference_rule;

    LinguisticGraph *ellipsis_graph, *test_nac_graph_discard, *test_nac_graph_keep;
};

BOOST_FIXTURE_TEST_SUITE(CompilerSuite, CompilerFixture)

BOOST_AUTO_TEST_CASE(CompileMatchIntoLinguisticGraphTest){
    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_ellipsis, oc::Compiler::Text, NONE));

    bool status = compiler->compile();

    BOOST_CHECK_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        LinguisticGraph *g = r->graph();

        BOOST_CHECK_EQUAL(g->nodesSize(), 4);
        BOOST_CHECK_EQUAL(g->edgesSize(), 3);

        //Check the structure and the features of the graph
        short count = 0;
        LinguisticGraph::const_ordered_node_iter nit, nit_end;
        for(boost::tie(nit, nit_end) = g->constOrderedNodes(); nit != nit_end; ++nit)
        {
            const LinguisticNodeProperties &node = g->nodeProperties(nit->first);
            if(count == 0)
            {
                BOOST_CHECK_EQUAL(g->nodeOutDegree(nit->first), 2);
                BOOST_CHECK_EQUAL(g->nodeInDegree(nit->first), 0);
                BOOST_CHECK_EQUAL(node.hyperpos(), "V");

                LinguisticGraph::out_edge_iter oit, oit_end;
                std::set<unsigned> possible_targets = ba::list_of(1)(2);
                for(boost::tie(oit, oit_end) = g->outEdges(nit->first); oit != oit_end; ++oit)
                {
                    const LinguisticNodeProperties &target = g->nodeProperties(g->targetNode(*oit));
                    BOOST_CHECK_EQUAL(possible_targets.find(target.id) != possible_targets.end(), true);
                }
            }
            else if(count == 1)
            {
                BOOST_CHECK_EQUAL(g->nodeOutDegree(nit->first), 0);
                BOOST_CHECK_EQUAL(g->nodeInDegree(nit->first), 1);
                BOOST_CHECK_EQUAL(node.hyperpos(), "");
                BOOST_CHECK_EQUAL(node.token(), "");
                BOOST_CHECK_EQUAL(node.lemma(), "");

                LinguisticGraph::in_edge_iter iit, iit_end;
                for(boost::tie(iit, iit_end) = g->inEdges(nit->first); iit != iit_end; ++iit)
                {
                    const LinguisticNodeProperties &source = g->nodeProperties(g->sourceNode(*iit));
                    const LinguisticEdgeProperties &edge   = g->edgeProperties(*iit);

                    BOOST_CHECK_EQUAL(source.id, 0);
                    BOOST_CHECK_EQUAL(edge.relFinal(), "suj");
                    BOOST_CHECK_EQUAL(edge.relCanonical(), "");
                }
            }
            else if(count == 2)
            {
                BOOST_CHECK_EQUAL(g->nodeOutDegree(nit->first), 1);
                BOOST_CHECK_EQUAL(g->nodeInDegree(nit->first), 1);
                BOOST_CHECK_EQUAL(node.hyperpos(), "C");
                BOOST_CHECK_EQUAL(node.token(), "");
                BOOST_CHECK_EQUAL(node.lemma(), "");

                LinguisticGraph::in_edge_iter iit, iit_end;
                for(boost::tie(iit, iit_end) = g->inEdges(nit->first); iit != iit_end; ++iit)
                {
                    const LinguisticNodeProperties &source = g->nodeProperties(g->sourceNode(*iit));
                    const LinguisticEdgeProperties &edge   = g->edgeProperties(*iit);

                    BOOST_CHECK_EQUAL(source.id, 0);
                    BOOST_CHECK_EQUAL(edge.relFinal(), "coord");
                    BOOST_CHECK_EQUAL(edge.relCanonical(), "");
                }

                LinguisticGraph::out_edge_iter oit, oit_end;
                for(boost::tie(oit, oit_end) = g->outEdges(nit->first); oit != oit_end; ++oit)
                {
                    const LinguisticNodeProperties &target = g->nodeProperties(g->targetNode(*oit));
                    const LinguisticEdgeProperties &edge   = g->edgeProperties(*oit);

                    BOOST_CHECK_EQUAL(target.id, 3);
                    BOOST_CHECK_EQUAL(edge.relFinal(), "dep.coord");
                    BOOST_CHECK_EQUAL(edge.relCanonical(), "");
                }

            }
            else if(count == 3)
            {
                BOOST_CHECK_EQUAL(g->nodeOutDegree(nit->first), 0);
                BOOST_CHECK_EQUAL(g->nodeInDegree(nit->first), 1);
                BOOST_CHECK_EQUAL(node.hyperpos(), "V");
                BOOST_CHECK_EQUAL(node.token(), "");
                BOOST_CHECK_EQUAL(node.lemma(), "");
            }

            count++;
        }

    }

    //delete compiler;
}

BOOST_AUTO_TEST_CASE(CompileCommandsTest){
    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_ellipsis_command, oc::Compiler::Text, NONE));

    bool status = compiler->compile();
    BOOST_REQUIRE_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        LinguisticGraph *g = r->graph();
        BOOST_CHECK_EQUAL(r->commandsCount(), 2);

        short ccount = 0;
        oc::Rule::CommandVectorConstIter cit, cit_end;
        for(boost::tie(cit, cit_end) = r->commands(); cit != cit_end; ++cit)
        {
            //Create matching between rule and graph
            std::map<unsigned, unsigned> nodes = ba::map_list_of(0, 1)(1, 0)(2,2)(3,3),
                    edges = ba::map_list_of(0,0)(1,1)(2,2);

            oc::Command *c = *cit;
            if(ccount == 0)
            {
                BOOST_CHECK_EQUAL(c->name(), "add_feature");

                //Apply command
                c->process(nodes, edges, ellipsis_graph);

                //mange should have a new feature void=y
                BOOST_CHECK_EQUAL(ellipsis_graph->nodeProperties(1).feature("void"), "y");
                //mange shouldn't have the feature test=n yet
                BOOST_CHECK_EQUAL(ellipsis_graph->nodeProperties(1).feature("test"), "");

            }else if(ccount == 1){
                BOOST_CHECK_EQUAL(c->name(), "add_feature");

                //Apply command
                c->process(nodes, edges, ellipsis_graph);

                //mange should have a new feature test=n
                BOOST_CHECK_EQUAL(ellipsis_graph->nodeProperties(1).feature("test"), "n");
                //mange should still have the feature void=y
                BOOST_CHECK_EQUAL(ellipsis_graph->nodeProperties(1).feature("void"), "y");
            }
            ccount++;
        }
    }

    //delete compiler;
}

BOOST_AUTO_TEST_CASE(CompileCommandWithReferenceTest)
{
    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_reference_rule, oc::Compiler::Text, NONE));

    bool status = compiler->compile();
    BOOST_REQUIRE_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        LinguisticGraph *g = r->graph();

        BOOST_CHECK_EQUAL(r->commandsCount(), 1);

        short ccount = 0;
        oc::Rule::CommandVectorConstIter cit, cit_end;
        for(boost::tie(cit, cit_end) = r->commands(); cit != cit_end; ++cit)
        {
            //Create matching between rule and graph
            std::map<unsigned, unsigned> nodes = ba::map_list_of(0, 3)(1, 2),
                    edges = ba::map_list_of(0,2);

            oc::Command *c = *cit;
            if(ccount == 0)
            {
                BOOST_CHECK_EQUAL(c->name(), "add_feature");
                //Apply command
                c->process(nodes, edges, test_nac_graph_keep);
                //aime SHOULD HAVE token=V
                BOOST_CHECK_EQUAL(test_nac_graph_keep->nodeProperties(3).feature("token"), "V");
            }
            ccount++;
        }
    }

    //delete compiler;
}

BOOST_AUTO_TEST_CASE(CompileStatusEdgeCommandTest){
    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_status_edge_command, oc::Compiler::Text, NONE));

    bool status = compiler->compile();
    BOOST_REQUIRE_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        LinguisticGraph *g = r->graph();
        BOOST_CHECK_EQUAL(r->commandsCount(), 1);

        oc::Rule::CommandVectorConstIter cit, cit_end;
        for(boost::tie(cit, cit_end) = r->commands(); cit != cit_end; ++cit)
        {
            oc::Command *c = *cit;
            BOOST_CHECK_EQUAL(c->name(), "edge_status");
        }
    }

    //delete compiler;
}

BOOST_AUTO_TEST_CASE(CompileConstraintsTest){
    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_ellipsis_constraint, oc::Compiler::Text, NONE));

    bool status = compiler->compile();
    BOOST_CHECK_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        LinguisticGraph *g = r->graph();
        BOOST_CHECK_EQUAL(r->edgesWithConstraintCount(), 1);

        //Create matching between rule and graph
        std::map<unsigned, unsigned> nodes = ba::map_list_of(0, 1)(1, 0)(2,2)(3,3),
                edges = ba::map_list_of(0,0)(1,1)(2,2);

        oc::Rule::ConstraintVectorConstIter cit, cit_end;
        for(boost::tie(cit, cit_end) = r->constraints(); cit != cit_end; ++cit)
        {
            oc::Rule::ConstraintIdType ids = (*cit).first;
            oc::Constraint *c = (*cit).second;

            std::pair<unsigned, unsigned> pair_ids =
                    boost::get< std::pair<unsigned, unsigned> >(ids);

            unsigned v = r->get_node_id("v").first, v2 = r->get_node_id("v2").first;
            LinguisticEdgeProperties cedge(ellipsis_graph->lastEdgeId()+1,
                                           "__ogre__constraint__", "__ogre__constraint__");
            ellipsis_graph->addEdge( nodes.find(v2)->second, nodes.find(v)->second, cedge );

            BOOST_REQUIRE_EQUAL(ellipsis_graph->edgesSize(), 4);

            //Apply constraint on newly created edge
            c->trigger(ellipsis_graph, ellipsis_graph->lastEdgeId());

            //Constraint should have added a new edge
            BOOST_CHECK_EQUAL(ellipsis_graph->edgesSize(), 5);

            //dort should have an out edge now
            BOOST_CHECK_EQUAL(ellipsis_graph->nodeOutDegree(3), 2);
            LinguisticGraph::out_edge_iter oit, oit_end;
            bool found = true;
            for(boost::tie(oit, oit_end) = ellipsis_graph->outEdges(3); oit != oit_end; ++oit)
            {
                const LinguisticEdgeProperties &edge = g->edgeProperties(*oit);
                if(edge.relFinal() == "suj" && edge.type() == LinguisticEdgeProperties::Deep && g->nodeProperties(g->targetNode(*oit)).lemma() == "Jean"){
                    found = true;
                }
            }
            BOOST_CHECK_EQUAL(found, true);
        }
    }
}

BOOST_AUTO_TEST_CASE(CompileGraphNacShouldKeepTheMatch){
    typedef vector<map<unsigned, unsigned> > MatchesType;
    typedef std::pair<MatchesType, MatchesType> MatchesPairType;

    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_nac_rule,
                                                               oc::Compiler::Text, DEBUG));

    bool status = compiler->compile();
    BOOST_CHECK_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        //NAC number should be equal to 1
        BOOST_CHECK_EQUAL(r->NACCount(), 1);

        GraphsMatcher m;
        MatchesPairType matching = m(r->graph(), test_nac_graph_keep);

        //There should be only 1 matching
        BOOST_CHECK_EQUAL(matching.first.size(), 1);
        BOOST_CHECK_EQUAL(matching.second.size(), 1);

        oc::Rule::NacVectorConstIter nac_begin, nac_end;
        for(boost::tie(nac_begin, nac_end) = r->NACs(); nac_begin != nac_end; ++nac_begin)
        {
            NAC *nac = *nac_begin;

            bool is_kept = false;
            if(nac){
                is_kept = nac->keep(test_nac_graph_keep, matching.first[0], matching.second[0]);
            }

            BOOST_CHECK_EQUAL(is_kept, true);
        }
    }
}

BOOST_AUTO_TEST_CASE(CompileGraphNacShouldDiscardTheMatch){
    typedef vector<map<unsigned, unsigned> > MatchesType;
    typedef std::pair<MatchesType, MatchesType> MatchesPairType;

    boost::shared_ptr<oc::Compiler> compiler (new oc::Compiler(grql_nac_rule,
                                                               oc::Compiler::Text, DEBUG));

    bool status = compiler->compile();
    BOOST_CHECK_EQUAL(status, true);

    oc::Compiler::RulesConstIter rit, rit_end;
    for(boost::tie(rit, rit_end) = compiler->rules(); rit != rit_end; ++rit)
    {
        oc::Rule *r = *rit;
        //NAC number should be equal to 1
        BOOST_CHECK_EQUAL(r->NACCount(), 1);

        GraphsMatcher m;
        MatchesPairType matching = m(r->graph(), test_nac_graph_discard);

        //There should be only 1 matching
        BOOST_CHECK_EQUAL(matching.first.size(), 1);
        BOOST_CHECK_EQUAL(matching.second.size(), 1);

        oc::Rule::NacVectorConstIter nac_begin, nac_end;
        for(boost::tie(nac_begin, nac_end) = r->NACs(); nac_begin != nac_end; ++nac_begin)
        {
            NAC *nac = *nac_begin;

            bool is_kept = true;
            if(nac){
                is_kept = nac->keep(test_nac_graph_discard, matching.first[0], matching.second[0]);
            }

            BOOST_CHECK_EQUAL(is_kept, false);
        }
    }
}

BOOST_AUTO_TEST_SUITE_END()
